#!/usr/bin/env python
# coding: utf-8
# test tag clusters


# test_capitalize.py

def capital_case(x):
    return x.capitalize()


def test_capital_case():
    assert capital_case('semaphore') == 'Semaphore'

# Emojitest
#n = '❤️👨‍⚕️'
#n = '😍,146'
# print(n.encode("utf-8"))
# n = '👨‍⚕️' #medical emoji with zero-width joiner (http://www.unicode.org/emoji/charts/emoji-zwj-sequences.html)
#nlist = Utils.extract_emojis(n)
# with open("emojifile.txt", "w", encoding='utf-8') as emojifile:
#    emojifile.write("Original: " + n + '\n')
#    for xstr in nlist:
#        emojifile.write('Emoji Extract: U+%04x' % ord(xstr) + '\n')
#        emojifile.write(xstr + '\n')
#    for _c in n:
#        emojifile.write(str(unicode_name(_c)) + '\n')
#        emojifile.write('Each Codepoint: U+%04x' % ord(_c) +  '\n')
# initialize global variables for analysis bounds (lat, lng coordinates)



#!/usr/bin/env python
# coding: utf-8
# test tag clusters

"""Emoji test module

This file is still on the TO_DO list
"""
# test_capitalize.py


def capital_case(x):
    return x.capitalize()


def test_capital_case():
    assert capital_case('semaphore') == 'Semaphore'

# Emojitest
# n = '❤️👨‍⚕️'
# n = '😍,146'
# n = '👨‍👩‍👦‍👦' (grapheme cluster)
# print(n.encode("utf-8"))
# n = '👨‍⚕️' #medical emoji with zero-width joiner
# see:
# http://www.unicode.org/emoji/charts/emoji-zwj-sequences.html)
# nlist = Utils.extract_emojis(n)
# with open("emojifile.txt", "w", encoding='utf-8') as emojifile:
#    emojifile.write("Original: " + n + '\n')
#    for xstr in nlist:
#        emojifile.write('Emoji Extract: U+%04x' % ord(xstr) + '\n')
#        emojifile.write(xstr + '\n')
#    for _c in n:
#        emojifile.write(str(unicode_name(_c)) + '\n')
#        emojifile.write('Each Codepoint: U+%04x' % ord(_c) +  '\n')
# initialize global variables for
# analysis bounds (lat, lng coordinates)

# see
# https://stackoverflow.com/questions/43852668/using-collections-counter-to-count-emojis-with-different-colors
# we want to ignore fitzpatrick modifiers
# and treat all differently colored emojis the same
# https://stackoverflow.com/questions/38100329/some-emojis-e-g-have-two-unicode-u-u2601-and-u-u2601-ufe0f-what-does
# COOKING
# OK HAND SIGN
# EMOJI MODIFIER FITZPATRICK TYPE-1-2
# GRINNING FACE WITH SMILING EYES
# HEAVY BLACK HEART
# WEARY CAT FACE
# SMILING FACE WITH HEART-SHAPED EYES
# OK HAND SIGN
# EMOJI MODIFIER FITZPATRICK TYPE-1-2
# GRINNING FACE WITH SMILING EYES
# PERSON WITH FOLDED HANDS
# EMOJI MODIFIER FITZPATRICK TYPE-3
# WEARY CAT FACE

# def cleanEmoji(c):
#    tuple = (u'\ufeff',u'\u200b',u'\u200d')
#    for ex in tuple:
#        c.replace(ex,"")
#    return(c)
# https://github.com/carpedm20/emoji/
# https://github.com/carpedm20/emoji/issues/75
# tc unicode problem
# https://stackoverflow.com/questions/40222971/python-find-equivalent-surrogate-pair-from-non-bmp-unicode-char

# @staticmethod
# def _extract_emoji_slow(string_with_emoji):
#     return [x.encode('utf-8') for x
#             in gc.grapheme_clusters(
#             string_with_emoji)]

# @staticmethod
# def _is_flag_emoji(c):
#     return "\U0001F1E6\U0001F1E8" <= c <= "\U0001F1FF\U0001F1FC" \
#            or c in [
#            "\U0001F3F4\U000e0067\U000e0062\U000e0065\U000e006e" \
#            "\U000e0067\U000e007f", "\U0001F3F4\U000e0067\U000e0062" \
#            "\U000e0073\U000e0063\U000e0074\U000e007f",
#            "\U0001F3F4\U000e0067\U000e0062\U000e0077\U000e006c" \
#            "\U000e0073\U000e007f"]

# enable for map display
# from mpl_toolkits.basemap import Basemap
# from PIL import Image
# https://gis.stackexchange.com/questions/127427/transforming-shapely-polygon-and-multipolygon-objects
# alternative Shapefile module pure Python
# https://github.com/GeospatialPython/pyshp#writing-shapefiles
# import shapefile
# to test:
# bba9f14180134069a9da5b3eb3539130822f917a55fc42d4adcc816018f79d9e "bei;bilzbad;flamingos;radebeul;@mcfitti;👯🎉😋;30grad;und","radebeul;bilzbad;flamingos","🐬;🌊;😋;🎉;👯"



"""
Run Integration test for TagMaps


This will check complete process of main() as if
evoked with tagmaps --autoMode --maxItems 50
"""

import sys
import os
import traceback
from pathlib import Path
from tagmaps.__main__ import main as tm_main


def tagmaps_system_integration_test():
    """Test complete method integration using main()"""
    # override resource path
    os.environ["TAGMAPS_RESOURCES"] = str(Path.cwd() / "resources")
    try:
        tm_main()
        sys.exit(0)
    except Exception:
        traceback.print_exc()
        sys.exit(1)


if __name__ == '__main__':
    tagmaps_system_integration_test()



# -*- coding: utf-8 -*-

"""
Module for shared structural elements
"""

from __future__ import absolute_import

from collections import namedtuple
from typing import List, Tuple

import attr

LOCATIONS: str = 'Locations'
TAGS: str = 'Tags'
EMOJI: str = 'Emoji'
TOPICS: List[str] = 'Topics'
ClusterType: Tuple[Tuple[str, int]] = (
    (LOCATIONS, 1),
    (TAGS, 2),
    (EMOJI, 3),
    (TOPICS, 4),
)

CleanedPost_ = namedtuple(  # pylint: disable=C0103
    'CleanedPostTuple',
    'origin_id lat lng guid user_guid '
    'post_create_date post_publish_date '
    'post_body hashtags emoji '
    'post_views_count post_like_count loc_id '
    'loc_name')


class CleanedPost(CleanedPost_):
    """This a wrapper class for namedtuple

    CleanedPost namedtuple provides access
    for fast handling
    of essential post attributes
    for use in tag maps clustering

    - namedtuple instances are just as memory
    efficient as regular tuples because they
    do not have per-instance dictionaries.
    """


@attr.s
class PostStructure(object):
    """Shared structure for Post Attributes

    Contains attributes shared among PG DB and LBSN ProtoBuf spec.
    - this could also be replaces by protobuf lbsnPost() from
    lbsnstructure package

    This list of arguments contains many  optional sets of information
    that can be adapted for specific cases / individual hooks
    """
    # origin of data (type int),
    # e.g. Flickr = 1, Twitter = 2
    origin_id = attr.ib(init=False)
    # global unique id of post (type str),
    # can be hashed or encrypted
    guid = attr.ib(init=False)
    # post latlng (type PostGis geometry.Point),
    # optional
    post_latlng = attr.ib(init=False)
    # global unique id (type str) for referenced place,
    # optional
    place_guid = attr.ib(init=False)
    # global unique id (type str) for referenced city,
    # optional
    city_guid = attr.ib(init=False)
    # global unique id (type str) for referenced country,
    # optional
    country_guid = attr.ib(init=False)
    # geoaccuracy of spatial information (type str),
    # either  'latlng', 'place', 'city' or 'country', optional
    post_geoaccuracy = attr.ib(init=False)
    # global unique id of user (type str),
    # can be hashed or encrypted
    user_guid = attr.ib(init=False)
    # the time the post was created,
    # e.g. the timestamp of the photo on Flickr
    # (type str, e.g. 2017-10-29 15:58:25), optional
    post_create_date = attr.ib(init=False)
    # the time the post was uploaded,
    # e.g. the timestamp of the upload on Flickr (type str),
    # optional
    post_publish_date = attr.ib(init=False)
    # content of post body,
    # e.g. the description on Flickr or the Tweet text on Twitter,
    # optional
    post_body = attr.ib(init=False)
    # language id (type str), optional
    post_language = attr.ib(init=False)
    # mentioned user_guids in post_body, optional
    user_mentions = attr.ib(init=False)
    # list of (hash-) tags (type str),
    # separated by semicolon, e.g. "lov;loved;awesomeday;goldenhour;"),
    # required or auto-extracted from post_body based on hashing '#'
    hashtags = attr.ib(init=False)
    # list of emoji (type str, separatior semicolon),
    # optional - can be auto-extracted from post_body
    emoji = attr.ib(init=False)
    # number of times this post has beed liked or
    # starred (Flickr), type int, optional
    post_like_count = attr.ib(init=False)
    # number of times this Post has been commented by other users,
    # e.g. count of Reply-Tweets in Twitter,
    # count of comments in Flickr etc., type int, optional
    post_comment_count = attr.ib(init=False)
    # number of times this post has beed viewed by other users,
    # type int, optional
    post_views_count = attr.ib(init=False)
    # the title of the post, type str, optional
    post_title = attr.ib(init=False)
    # thumbnail url, optional
    post_thumbnail_url = attr.ib(init=False)
    # post url, optional
    post_url = attr.ib(init=False)
    # post type, either 'text', 'image' or 'video',
    # type str, optional
    post_type = attr.ib(init=False)
    # applied post filters, type str, optional
    post_filter = attr.ib(init=False)
    # Number of times this Post has been quoted by other users,
    # e.g. count of Quote-Tweets in Twitter, type int, optional
    post_quote_count = attr.ib(init=False)
    # Number of times this Post has been shared by other users,
    # e.g. count of Retweets in Twitter, type int, optional
    post_share_count = attr.ib(init=False)
    # Type of input device used by the user to post,
    # for a list see Twitter, e.g. "Web", "IPhone",
    # "Android" etc., type str, optional
    input_source = attr.ib(init=False)
    # An integer for specifying licenses attached to post
    # (e.g. All Rights Reserved = 0). Numbers shamelessly
    # ripped from Flickr:
    # https://www.flickr.com/services/api/flickr.photos.licenses.getInfo.html
    # type int, optional
    post_content_license = attr.ib(init=False)
    # latitude of post in decimal degrees (WGS 1984),
    # type float, required
    latitude = attr.ib(init=False)
    # longitude of post in decimal degrees  (WGS 1984),
    # type float, required
    longitude = attr.ib(init=False)
    # location_id, automatically
    # constructed from lat:lng
    loc_id = attr.ib(init=False)
    # name of location, type str, optional
    loc_name = attr.ib(init=False)


class AnalysisBounds():
    """Structure for storing boundary (lim lat/lng)"""

    def __init__(self):
        """initialize global variables for analysis bounds

        (lat, lng coordinates)
        """
        self.lim_lat_min = None
        self.lim_lat_max = None
        self.lim_lng_min = None
        self.lim_lng_max = None

    def upd_latlng_bounds(self, lat, lng):
        """Update lat/lng bounds based on coordinate pair."""
        if self.lim_lat_min is None or \
                (lat < self.lim_lat_min and not lat == 0):
            self.lim_lat_min = lat
        if self.lim_lat_max is None or \
                (lat > self.lim_lat_max and not lat == 0):
            self.lim_lat_max = lat
        if self.lim_lng_min is None or \
                (lng < self.lim_lng_min and not lng == 0):
            self.lim_lng_min = lng
        if self.lim_lng_max is None or \
                (lng > self.lim_lng_max and not lng == 0):
            self.lim_lng_max = lng

    def get_bound_report(self):
        """Report on spatial bounds"""
        if (self.lim_lat_min is None
                or self.lim_lat_max is None
                or self.lim_lng_min is None
                or self.lim_lng_max is None):
            return f'Bounds have not been initlialized'
        bound_report = f'Bounds are: ' \
            f'Min {float(self.lim_lng_min)} {float(self.lim_lat_min)} ' \
            f'Max {float(self.lim_lng_max)} {float(self.lim_lat_max)}'
        return bound_report



# -*- coding: utf-8 -*-

"""
Module for tag maps alpha shapes generation
"""

from __future__ import absolute_import

import math
from collections import namedtuple

import numpy as np

from scipy.spatial import Delaunay  # pylint: disable=E0611
import shapely.geometry as geometry
from shapely.ops import cascaded_union, polygonize

from tagmaps.classes.compile_output import Compile

# results will be compiled as namedtuple
AlphaShapesAndMeta = namedtuple(
    'AlphaShapesAndMeta',
    'shape post_count views user_count item_name item_totalcount '
    'weightsv1 weightsv2 weightsv3 shapetype')


class AlphaShapes():
    """Converts (cluster) point clouds to shapes"""
    @staticmethod
    def get_single_cluster_shape(
            item, single_post,
            cluster_distance: float, proj_coords):
        """Get Shapes for items with no clusters
        Will return a buffer based on cluster distance
        """
        shapetype = "Single cluster"
        # project lat/lng to UTM
        point_x, point_y = proj_coords(
            single_post.lng, single_post.lat)
        pcoordinate = geometry.Point(point_x, point_y)
        # single dots are presented
        # as buffers with 0.5% of width-area
        result_polygon = pcoordinate.buffer(
            cluster_distance/4,
            resolution=3)
        # result_polygon = pcoordinate.buffer(
        #   min(distXLng,distYLat)/100,
        #   resolution=3)
        if (result_polygon is None or
                result_polygon.is_empty):
            return None
        # append statistics for item with no cluster
        single_cluster_result = AlphaShapesAndMeta(
            result_polygon, 1, max(single_post.post_views_count,
                                   single_post.post_like_count), 1, str(item[0]),
            item[1], 1, 1, 1, shapetype
        )
        return single_cluster_result

    @staticmethod
    def get_cluster_shape(
            item, clustered_post_guids,
            cleaned_post_dict,
            cluster_distance: float,
            local_saturation_check,
            proj_coords):
        """Returns alpha shapes and tag_area (sqm) for a point cloud"""
        # we define a new list of Temp Alpha Shapes outside the loop,
        # so that it is not overwritten each time
        alphashapes_and_meta_tmp = list()
        item_area = 0

        for post_guids in clustered_post_guids:
            # for each cluster for this toptag
            posts = [cleaned_post_dict[x]
                     for x in post_guids]
            post_count = len(post_guids)
            unique_user_count = len(
                {post.user_guid for post in posts})
            sum_views = sum(
                [post.post_views_count for post in posts])
            # needs to be moved to CompileOutput:
            weightsv1 = Compile.get_weight(
                1, post_count, unique_user_count)

            weightsv2 = Compile.get_weight(
                2, post_count, unique_user_count)
            weightsv3 = Compile.get_weight(
                3, post_count, unique_user_count)
            distinct_locations = {post.loc_id
                                  for post in posts}
            # simple list comprehension with projection:
            points = [geometry.Point(
                proj_coords(
                    float(location.split(':')[1]), float(location.split(':')[0]))
            ) for location in distinct_locations]

            # get poly shape from points
            result = AlphaShapes._get_poly(points, cluster_distance)
            result_polygon = result[0]
            shapetype = result[1]

            # Geom, Join_Count, Views,  COUNT_User,ImpTag,TagCountG,HImpTag
            if result_polygon is not None and not result_polygon.is_empty:
                if local_saturation_check:
                    item_area += result_polygon.area
                alphashapedata = AlphaShapesAndMeta(
                    result_polygon, post_count, sum_views,
                    unique_user_count, item[0], item[1],
                    weightsv1, weightsv2, weightsv3,
                    shapetype)
                alphashapes_and_meta_tmp.append(alphashapedata)
        if alphashapes_and_meta_tmp:
            # finally sort and append all
            # cluster shapes for this tag
            alphashapes_and_meta_tmp = sorted(
                alphashapes_and_meta_tmp, key=lambda x: -x[6])
        AlphaShapesArea = namedtuple(
            'AlphaShapesArea', 'alphashape item_area')
        return AlphaShapesArea(alphashapes_and_meta_tmp, item_area)

    @staticmethod
    def _get_poly_five_to_ten(
            point_collection, cluster_distance):
        """ convex hull for small point clouds"""
        result_polygon = point_collection.convex_hull
        result_polygon = result_polygon.buffer(
            cluster_distance/4, resolution=3)
        shapetype = "between 5 and 10 points_convexHull"
        # result_polygon = result_polygon.buffer(
        #   min(distXLng,distYLat)/100,resolution=3)
        return result_polygon, shapetype

    @staticmethod
    def _get_poly_two_to_five(
            point_collection, cluster_distance):
        """ convex hull for tiny point clouds"""
        shapetype = "between 2 and 5 points_buffer"
        # calc distance between points, see
        # http://www.mathwarehouse.com/algebra/distance_formula/index.php
        # bdist = math.sqrt(
        #   (points[0].coords.xy[0][0]-points[1].coords.xy[0][0])**2
        # + (points[0].coords.xy[1][0]-points[1].coords.xy[1][0])**2)
        # print(str(bdist))
        # single dots are presented as buffer with 0.5% of width-area:
        poly_shape = point_collection.buffer(
            cluster_distance/4, resolution=3)
        poly_shape = poly_shape.convex_hull
        # result_polygon = point_collection.buffer(
        #   min(distXLng,distYLat)/100,resolution=3)
        # #single dots are presented as buffer with 0.5% of width-area
        return poly_shape, shapetype

    @staticmethod
    def _get_poly_larger_ten(
            points, pointcount,
            cluster_distance,
            point_collection):
        """ Get poly for larger point clouds
        This is still recursive trial & error

        Repeat generating alpha shapes with
        smaller alpha value if Multigon is
        generated. Smaller alpha values mean
        less granularity of resulting polygon,
        but too large alpha may result
        in empty polygon
        (this branch is sometimes
        executed for larger scales)
        """
        if pointcount > 500:
            startalpha = 1000000
        elif pointcount > 200:
            startalpha = 10000
        else:
            startalpha = 9000
        # concave hull/alpha shape /50000:
        poly_shape = AlphaShapes.alpha_shape(
            points, alpha=cluster_distance/startalpha)
        shapetype = "Initial Alpha Shape + Buffer"
        if not isinstance(poly_shape, bool) and poly_shape.is_empty:
            # try again with centered axis reduced alpha
            poly_shape = AlphaShapes.alpha_shape(
                points, alpha=(cluster_distance/startalpha), centeraxis=True)
        # if not poly_shape.is_empty:
        # print("Success")
        # check type comparison here
        if isinstance(poly_shape, (bool, geometry.multipolygon.MultiPolygon)):
            for i in range(1, 6):
                # try decreasing alpha
                # ** means cube
                alpha = startalpha + (startalpha * (i**i))
                # /100000
                poly_shape = AlphaShapes.alpha_shape(
                    points, alpha=cluster_distance/alpha)
                if not (isinstance(poly_shape,
                                   geometry.multipolygon.MultiPolygon)
                        and not isinstance(poly_shape, bool)):
                    shapetype = "Multipolygon Alpha Shape /" + \
                        str(alpha)
                    break
            if (isinstance(
                    poly_shape, (bool, geometry.multipolygon.MultiPolygon))):
                # try increasing alpha
                for i in range(1, 6):
                    # try decreasing alpha
                    alpha = startalpha / (i*i)
                    # /100000
                    poly_shape = AlphaShapes.alpha_shape(
                        points, alpha=cluster_distance/alpha)
                    if not (isinstance(
                            poly_shape, geometry.multipolygon.MultiPolygon)
                            and not isinstance(poly_shape, bool)):
                        shapetype = "Multipolygon Alpha Shape /" + \
                            str(alpha)
                        break
            if isinstance(poly_shape, geometry.multipolygon.MultiPolygon):
                shapetype = "Multipolygon Alpha Shape -> Convex Hull"
                # if still of type multipolygon,
                # try to remove holes and do a convex_hull
                poly_shape = poly_shape.convex_hull
            # OR: in case there was a problem with generating
            # alpha shapes
            # (circum_r = a*b*c/(4.0*area)
            # --> ZeroDivisionError: float division by zero)
            # this branch is rarely executed
            # for large point clusters where
            # alpha is perhaps set too small
        if (isinstance(poly_shape, bool)
                or poly_shape.is_empty or
                AlphaShapes._few_polyinpoints(points, poly_shape)):
            shapetype = ("BoolAlpha/Empty -> Fallback "
                         "to PointCloud Convex Hull")
            # convex hull
            poly_shape = point_collection.convex_hull
        # Finally do a buffer to smooth alpha
        poly_shape = poly_shape.buffer(
            cluster_distance/4, resolution=3)
        # result_polygon = result_polygon.buffer(
        #   min(distXLng,distYLat)/100,resolution=3)
        return poly_shape, shapetype

    @staticmethod
    def _few_polyinpoints(points, poly):
        """Check how many of original points are in Alpha

        Because some QHull Delauney return a small Alpha Shape that
        does not include most points, this is a fallback solution.
        Better solution would be to revise QHull Delauney/Alpha,
        but even with centered axis, this issue remains (see alpha_shape).
        """
        p_count = 1
        for point in points:
            if point.within(poly):
                p_count += 1
        perc = p_count/(len(points)/100)
        if perc < 50:
            return True
        return False

    @staticmethod
    def _get_poly_one(
            point_collection, cluster_distance):
        shapetype = "1 point cluster"
        # single dots are presented as buffer
        # with 0.5% of width-area
        poly_shape = point_collection.buffer(
            cluster_distance/4, resolution=3)
        # result_polygon = point_collection.buffer(
        # min(distXLng,distYLat)/100,resolution=3)
        # #single dots are presented as buffer
        # with 0.5% of width-area
        return poly_shape, shapetype

    @staticmethod
    def _get_poly(points, cluster_distance):
        """Get polygon/shape for collection
        of geooordinates (points)

        Returns:
        Shape (Polygon)
        Shapetype (Approach of Shape-Gen)
        """
        point_collection = geometry.MultiPoint(
            list(points))
        poly_shape = None
        pointcount = len(points)
        if pointcount >= 5:
            if pointcount < 10:
                poly_shape, shapetype = AlphaShapes._get_poly_five_to_ten(
                    point_collection, cluster_distance)
            else:
                poly_shape, shapetype = AlphaShapes._get_poly_larger_ten(
                    points, pointcount, cluster_distance, point_collection)
        elif 2 <= pointcount < 5:
            poly_shape, shapetype = AlphaShapes._get_poly_two_to_five(
                point_collection, cluster_distance)
        elif (pointcount == 1 or
              isinstance(poly_shape, geometry.point.Point)
              or poly_shape is None):
            poly_shape, shapetype = AlphaShapes._get_poly_one(
                point_collection, cluster_distance)
        # final check for multipolygon
        if (isinstance(poly_shape, geometry.multipolygon.MultiPolygon)
                or poly_shape.is_empty):
            # usually not executed
            poly_shape = poly_shape.convex_hull
            shapetype = "Convex Hull Final Fallback"

        return poly_shape, shapetype

    @staticmethod
    def alpha_shape(points, alpha, centeraxis=None):
        """
        Alpha Shapes Code by KEVIN DWYER, see
        http://blog.thehumangeo.com/2014/05/12/drawing-boundaries-in-python/
            Compute the alpha shape (concave hull) of a set
            of points.
            @param points: Iterable container of points.
            @param alpha: alpha value to influence the
                gooeyness of the border. Smaller numbers
                don't fall inward as much as larger numbers.
                Too large, and you lose everything!

        with minor adaptions to Tag Maps clustering.

        Notes:
        see also floating point resolution issue:
        https://stackoverflow.com/questions/8071382/points-left-out-when-nearby-in-scipy-spatial-delaunay

        Consider replace with shapely.Delauney
        http://kuanbutts.com/2017/08/17/delaunay-triangulation/
        """
        if centeraxis is None:
            centeraxis = False
        if len(points) < 4:
            # When you have a triangle,
            # there is no sense
            # in computing an alpha shape.
            return geometry.MultiPoint(
                list(points)).convex_hull

        def add_edge(edges, edge_points, coords, i, j):
            """
            Add a line between the i-th and j-th points,
            if not in the list already
            """
            if (i, j) in edges or (j, i) in edges:
                # already added
                return
            edges.add((i, j))
            edge_points.append(coords[[i, j]])
        coords = np.array([point.coords[0]
                           for point in points])

        # qhull_options = 'Qt'
        # #To avoid this error,
        # joggle the data by specifying the
        # 'QJ' option to the DELAUNAY function.see:
        # https://de.mathworks.com/matlabcentral/answers/94438-why-does-the-delaunay-function-in-matlab-7-0-r14-produce-an-error-when-passed-colinear-points?s_tid=gn_loc_drop

        if centeraxis:
            # Center axis for avoiding floating point precision issues
            norm_co = coords[0]
            coords = coords - norm_co
        tri = Delaunay(coords)
        # print problematic points:
        # print(tri.coplanar)
        # tri = Delaunay(coords,qhull_topions='QJ')
        # #Version 3.1 added triangulated output ('Qt').
        # It should be used for Delaunay triangulations
        # instead of using joggled input ('QJ').
        edges = set()
        edge_points = []
        # loop over triangles:
        # vert_ia, ib, ic = indices of corner points of the
        # triangle
        for vert_ia, vert_ib, vert_ic in tri.vertices:
            pa_v = coords[vert_ia]
            pb_v = coords[vert_ib]
            pc_v = coords[vert_ic]
            # Lengths of sides of triangle
            a_val = math.sqrt((pa_v[0]-pb_v[0])**2 + (pa_v[1]-pb_v[1])**2)
            b_val = math.sqrt((pb_v[0]-pc_v[0])**2 + (pb_v[1]-pc_v[1])**2)
            c_val = math.sqrt((pc_v[0]-pa_v[0])**2 + (pc_v[1]-pa_v[1])**2)
            # Semiperimeter of triangle
            s_res = (a_val + b_val + c_val)/2.0
            # Area of triangle by Heron's formula
            try:
                area = math.sqrt(
                    s_res*(s_res-a_val)*(s_res-b_val)*(s_res-c_val))
            except ValueError:
                return False
            if area == 0:
                return False
            circum_r = a_val*b_val*c_val/(4.0*area)
            # Here's the radius filter.
            # print circum_r
            if circum_r < 1.0/alpha:
                add_edge(edges, edge_points, coords, vert_ia, vert_ib)
                add_edge(edges, edge_points, coords, vert_ib, vert_ic)
                add_edge(edges, edge_points, coords, vert_ic, vert_ia)

        if centeraxis:
            # return to original axis center
            edge_points = [np.array(
                [edgepoint[0]+norm_co, edgepoint[1]+norm_co]
            ) for edgepoint in edge_points]

        mmulti_line_str = geometry.MultiLineString(edge_points)
        triangles = list(polygonize(mmulti_line_str))
        return cascaded_union(triangles)  # , edge_points
        # return geometry.polygon.asPolygon(edge_points,holes=None)



# -*- coding: utf-8 -*-

"""
Config processing for tag maps package.
"""

from __future__ import absolute_import

import argparse
import configparser
import csv
import logging
import os
import warnings
from pathlib import Path
from typing import List, Set, Dict, Tuple

import fiona
import pyproj
from shapely.geometry import shape

from tagmaps import __version__
from tagmaps.classes.utils import Utils


class BaseConfig:
    """Optional class for handling of typical config parameters"""

    def __init__(self):
        # Set Default Config options here
        # or define options as input args
        self.data_source = "lbsn"
        self.cluster_tags = True
        self.cluster_locations = True
        self.epsg = True
        self.remove_long_tail = True
        self.cluster_emoji = True
        self.topic_modeling = False
        self.write_cleaned_data = True
        self.local_saturation_check = False
        self.tokenize_japanese = False  # currently not implemented
        self.shapefile_intersect = None
        self.shapefile_exclude = None
        self.ignore_stoplists = False
        self.selectionlist_emoji = None
        self.selectionlist_tags = None
        self.stoplist_emoji = None
        self.stoplist_tags = None
        self.stoplist_user = None
        self.stoplist_places = None
        self.ignore_place_corrections = False
        self.statistics_only = False
        self.limit_bottom_user_count = 5
        self.write_gis_comp_line = True
        self.auto_mode = False
        self.max_items = 1000
        self.filter_origin = None
        self.logging_level = logging.INFO
        self.input_folder = None
        self.output_folder = None
        self.config_folder = None
        self.cluster_cut_distance = None

        # additional auto settings
        self.sort_out_always_set = set()
        self.sort_out_always_instr_set = set()
        self.override_crs = None
        self.crs_proj = None
        self.epsg_code = ""
        self.sort_out_places_set = set()
        self.sort_out_user_set = None
        self.sort_out_places = False
        self.sort_out_emoji_set = None
        self.select_emoji_set = None
        self.select_tags_set = None
        self.correct_places = False
        self.correct_place_latlng_dict = dict()
        self.shp_geom = None
        self.shp_exclude_geom = None

        # initialization
        resource_path = os.environ.get("TAGMAPS_RESOURCES")

        self.pathname = Path.cwd()
        if resource_path:
            self.resource_path = Path(resource_path)
        else:
            self.resource_path = self.pathname
        self.load_from_intermediate = None
        self.parse_args()

        # set logger
        self.log = Utils.set_logger(self.output_folder, self.logging_level)

        if not self.load_from_intermediate and not self.input_folder.exists():
            raise ValueError(f"Folder {self.input_folder} not found.")
        if not self.config_folder.exists():
            raise ValueError(f"Folder {self.config_folder} not found.")

        self.load_filterlists()
        if self.shapefile_intersect:
            self.shp_geom = self.load_shapefile(self.shapefile_intersect)
        if self.shapefile_exclude:
            self.shp_exclude_geom = self.load_shapefile(self.shapefile_exclude)
        self.source_map = self.load_sourcemapping()
        self.set_glob_options()

    def parse_args(self):
        """Parse init args and set default values

        Note: All paths are relative. If you wish to use another path entirely,
        set TAGMAPS_RESOURCES environment variable to the absolute path to use
        """
        parser = argparse.ArgumentParser()
        parser.add_argument('--version',
                            action='version',
                            version=f'tagmaps {__version__}')
        parser.add_argument("-s",
                            "--source",
                            default="lbsn",
                            help="Specify type of source of data. "
                            "This is needed to read data. "
                            "Defaults to 'lbsn'.")
        parser.add_argument("-r",
                            "--disable_remove_longtail",
                            action="store_true",
                            help="This will disable exclusion "
                            "of tags that are used by only a "
                            "small number of users")
        parser.add_argument("-e",
                            "--epsg",
                            help="If provided, will overwrite "
                            "auto EPSG code")
        parser.add_argument("-t", "--disable_cluster_tags",
                            action="store_true",
                            help="Disable cluster tag locations"
                            "No clusters will be generated per "
                            "distinct tag.")
        parser.add_argument("-p",
                            "--disable_cluster_locations",
                            action="store_true",
                            help="Disable cluster of post locations "
                            )
        parser.add_argument("-c",
                            "--local_saturation_check",
                            action="store_true",
                            help="Will attempt to exclude any tags that "
                            "are (over)used at above a certain percentage "
                            "of locations in processing extent. This "
                            "may improve legibility of tag maps at "
                            "larger scales"
                            )
        parser.add_argument("-o",
                            "--disable_cluster_emoji",
                            action="store_true",
                            help="Disable cluster of "
                            "emoji in addition to tags.")
        parser.add_argument("-m",
                            "--topic_modeling",
                            action="store_true",
                            help="This will used topic modeling "
                            "to detect groups of tags (based on "
                            "Latent Dirichlet Allocation). "
                            "[Not implemented]")
        parser.add_argument("-w",
                            "--disable_write_cleaneddata",
                            action="store_true",
                            help="This will disable "
                            "write out of a file limited to the data "
                            "that will be used for tag clustering "
                            "(filtered based on stoplists, clipped boundary, "
                            "minimum user threshold etc.)"
                            )
        parser.add_argument("-i",
                            "--shapefile_intersect",
                            type=Path,
                            help="Provide a relative path to a shapefile "
                            "to clip data prior to clustering.")
        parser.add_argument("--shapefile_exclude",
                            type=Path,
                            help="Provide a relative path to a shapefile "
                            "to exclude data prior to clustering. If "
                            "--shapefile_intersect is also used, the exclusion "
                            "here applies after inclusion.")
        parser.add_argument("-n",
                            "--ignore_stoplists",
                            action="store_true",
                            help="Set this flag to ignore any stoplists, "
                            "even if available/supplied."
                            )
        parser.add_argument("--stoplist_emoji",
                            type=Path,
                            help="Supply a relative path to a text file "
                            "containing emoji to ignore "
                            "during emoji clustering. "
                            "Format: one entry per line."
                            )
        parser.add_argument("--stoplist_tags",
                            type=Path,
                            help="Supply a relative path to a text file "
                            "containing terms to ignore during tag clustering. "
                            "This list is supplemental to "
                            "00_Config/SortOutAlways.txt, if such file exists. "
                            "Format: one entry per line."
                            )
        parser.add_argument("--stoplist_user",
                            type=Path,
                            help="Supply a relative path to a text file "
                            "containing user (guids) to ignore during "
                            "tag clustering. Format: one entry per line."
                            )
        parser.add_argument("--stoplist_places",
                            type=Path,
                            help="Supply a relative path to a text file "
                            "containing places (guids) to ignore during tag clustering. "
                            "This list is supplemental to "
                            "00_Config/SortOutPlaces.txt, if such file exists. "
                            "Format: one entry per line."
                            )
        parser.add_argument("--selectionlist_emoji",
                            type=Path,
                            help="Supply a relative path to a text file "
                            "containing emoji, as a positive filter list "
                            "for selecting a subset of emoji to focus on. "
                            "Format: one entry per line."
                            )
        parser.add_argument("--selectionlist_tags",
                            type=Path,
                            help="Supply a relative path to a text file "
                            "containing tags, as a positive filter list "
                            "for selecting a subset of tags to focus on. "
                            "Format: one entry per line."
                            )
        parser.add_argument("-b",
                            "--ignore_place_corrections",
                            action="store_true",
                            help="Set this flag to ignore place corrections, "
                            "even if available, supplied."
                            )
        parser.add_argument("--statistics_only",
                            action="store_true",
                            help="Do not cluster, only read input data"
                            " and output statistics. Suitable to get an "
                            "overview of input data."
                            )
        parser.add_argument("-d", "--cluster_cut_distance",
                            help="Provide a cluster cut distance (in meters) "
                            "where the clustering will be stopped. This "
                            "will override the auto detection "
                            "of cluster distance.",
                            type=float)
        parser.add_argument("-l",
                            "--limit_bottom_usercount",
                            type=int,
                            default=5,
                            help="Remove all tags that are used by "
                            "less than x photographers.")
        parser.add_argument("-g",
                            "--disable_write_gis_comp_line",
                            action="store_true",
                            default=True,
                            help="Disables writing placeholder "
                            "entry after headerline "
                            "for avoiding GIS import format issues. "
                            "Some GIS programs will decide format of "
                            "columns based on the first data line in a CSV. "
                            "This setting makes sure there're no empty fields "
                            "in the first line.",
                            )
        parser.add_argument("-a",
                            "--auto_mode",
                            action="store_true",
                            help="If set, no user input will "
                            "be requested during processing. "
                            "This flag is suitable if you're running in a "
                            "command line only environment (WSL) that "
                            "cannot use the tKinter GUI.",
                            )
        parser.add_argument("-f",
                            "--filter_origin",
                            type=str,
                            help="If provided, will filter input data "
                            "based on origin_id column and the value "
                            "provided.",
                            )
        parser.add_argument("-v", "--verbose",
                            help="Increase output verbosity",
                            action="store_true")
        parser.add_argument("-x", "--max_items",
                            help="Number of distinct items to process",
                            default=1000,
                            type=int)
        parser.add_argument("-q", "--output_folder",
                            help="Relative path to output folder",
                            default="02_Output",
                            type=Path)
        parser.add_argument("-j", "--input_folder",
                            help="Relative path to input folder",
                            default="01_Input",
                            type=Path)
        parser.add_argument("-k", "--config_folder",
                            help="Relative path to config folder",
                            default="00_Config",
                            type=Path)
        parser.add_argument("-u", "--load_intermediate",
                            help="A (relative) path to load from "
                            "intermediate (cleaned) data",
                            type=Path)

        args = parser.parse_args()
        if args.verbose:
            self.logging_level = logging.DEBUG
        if args.source:
            self.data_source = args.source
        if args.disable_cluster_tags:
            self.cluster_tags = False
        if args.disable_cluster_locations:
            self.cluster_locations = False
        if args.epsg is None:
            self.override_crs = None
        else:
            self.load_custom_crs(self.epsg)
        if args.disable_remove_longtail:
            self.remove_long_tail = False
        if args.disable_cluster_emoji:
            self.cluster_emoji = False
        if args.topic_modeling:
            self.topic_modeling = True
        if args.disable_write_cleaneddata:
            self.write_cleaned_data = False
        if args.local_saturation_check:
            self.local_saturation_check = True
        if args.shapefile_intersect:
            self.shapefile_intersect = \
                self.resource_path / args.shapefile_intersect
        if args.shapefile_exclude:
            self.shapefile_exclude = \
                self.resource_path / args.shapefile_exclude
        if args.ignore_stoplists:
            self.ignore_stoplists = True
        if args.selectionlist_emoji:
            self.selectionlist_emoji = \
                self.resource_path / args.selectionlist_emoji
        if args.selectionlist_tags:
            self.selectionlist_tags = \
                self.resource_path / args.selectionlist_tags
        if args.stoplist_emoji:
            self.stoplist_emoji = self.resource_path / args.stoplist_emoji
        if args.stoplist_tags:
            self.stoplist_tags = self.resource_path / args.stoplist_tags
        if args.stoplist_user:
            self.stoplist_user = self.resource_path / args.stoplist_user
        if args.stoplist_places:
            self.stoplist_places = self.resource_path / args.stoplist_places
        if args.ignore_place_corrections:
            self.ignore_place_corrections = True
        if args.statistics_only:
            self.statistics_only = True
        if args.limit_bottom_usercount:
            self.limit_bottom_user_count = args.limit_bottom_usercount
        if args.disable_write_gis_comp_line:
            self.write_gis_comp_line = False
        if args.auto_mode:
            self.auto_mode = args.auto_mode
        if args.filter_origin:
            self.filter_origin = args.filter_origin
        if args.max_items:
            self.max_items = args.max_items
        if args.output_folder:
            self.output_folder = self.resource_path / args.output_folder
        if args.input_folder:
            self.input_folder = self.resource_path / args.input_folder
        if args.config_folder:
            self.config_folder = self.resource_path / args.config_folder
        if args.load_intermediate:
            self.load_from_intermediate = self.resource_path / args.load_intermediate
        if args.cluster_cut_distance:
            self.cluster_cut_distance = args.cluster_cut_distance

    def load_filterlists(self):
        """Load filterlists for filtering terms (in-string and full match)
        and places, including place lat/lng corrections.
        """
        # locations for files
        sort_out_always_file = (
            self.config_folder / "SortOutAlways.txt")
        sort_out_always_instr_file = (
            self.config_folder / "SortOutAlways_inStr.txt")
        sort_out_places_file = (
            self.config_folder / "SortOutPlaces.txt")
        correct_place_latlng_file = (
            self.config_folder / "CorrectPlaceLatLng.txt")
        # load lists
        self.sort_out_always_set = self.load_filterlist(sort_out_always_file)
        if self.stoplist_tags:
            self.sort_out_always_set.update(
                self.load_filterlist(self.stoplist_tags)
            )
        self.sort_out_always_instr_set = self.load_filterlist(
            sort_out_always_instr_file)
        self.sort_out_places_set = self.load_place_stoplist(
            sort_out_places_file)
        if self.stoplist_places:
            self.sort_out_places_set.update(
                self.load_filterlist(self.stoplist_places)
            )
        if self.stoplist_user:
            self.sort_out_user_set = self.load_filterlist(
                self.stoplist_user)
        if self.stoplist_emoji:
            self.sort_out_emoji_set = self.load_filterlist(
                self.stoplist_emoji)
        if self.selectionlist_emoji:
            self.select_emoji_set = self.load_filterlist(
                self.selectionlist_emoji)
        if self.selectionlist_tags:
            self.select_tags_set = self.load_filterlist(
                self.selectionlist_tags)
        self.correct_place_latlng_dict = self.load_place_corrections(
            correct_place_latlng_file
        )
        # log results
        self._report_loaded_lists()

    def _report_loaded_lists(self):
        """Report stats of loaded stop & selection lists"""
        Utils.report_listload(
            self.sort_out_always_set, "stoplist items")
        Utils.report_listload(
            self.sort_out_always_instr_set, "inStr stoplist items")
        Utils.report_listload(
            self.sort_out_places_set, "stoplist places")
        Utils.report_listload(
            self.stoplist_places, "stoplist places")
        Utils.report_listload(
            self.correct_place_latlng_dict, "place lat/lng corrections")
        Utils.report_listload(
            self.sort_out_user_set, "user stoplist items")
        Utils.report_listload(
            self.select_emoji_set, "emoji selection items")
        Utils.report_listload(
            self.select_tags_set, "tags selection items")

    def load_filterlist(self, file: Path) -> Set[str]:
        """Loads filterlist of terms from file returns set"""
        if self.ignore_stoplists is True:
            return
        if not os.path.isfile(file):
            self.log.warning(f"{file} not found.")
            return
        store_set = set()
        with open(file, newline="", encoding="utf8") as f_handle:
            store_set = {line.lower().rstrip("\r\n") for line in f_handle}
        return store_set

    def load_place_stoplist(self, file):
        """Loads stoplist places from file and stores in set"""
        if not os.path.isfile(file) or self.ignore_stoplists is True:
            return
        store_set = set()
        with open(file, newline="", encoding="utf8") as f_handle:
            f_handle.readline()
            # Get placeid
            store_set = {
                line.rstrip("\r\n").split(",")[0]
                for line in f_handle if len(line) > 0}
        self.sort_out_places = True
        return store_set

    def load_place_corrections(
            self, file: Path) -> Dict[str, Tuple[float, float]]:
        """Fills dictionary with list of corrected lat/lng entries
        e.g.: Dictionary: placeid = lat, lng

        - sets self.correct_places to True
        """
        store_dict = dict()
        if os.path.isfile(file) or self.ignore_place_corrections is True:
            return
        with open(file, newline="", encoding="utf8") as f_handle:
            f_handle.readline()
            for line in f_handle:
                if not line:
                    continue
                linesplit = line.rstrip("\r\n").split(",")
                if len(linesplit) == 1:
                    continue
                store_dict[linesplit[0]] = (linesplit[1], linesplit[2])
        self.correct_places = True
        return store_dict

    def load_shapefile(self, shapefile_path: Path) -> List[shape]:
        """Imports single or multi polygon shapefile for intersecting points"""
        poly_shape = fiona.open(shapefile_path)
        vertices_count = 0
        shp_geom = []
        if len(poly_shape) == 1:
            # simple polygon
            geom = next(iter(poly_shape))
            vertices_count = len(geom["geometry"]["coordinates"][0])
            shp_geom.append(shape(geom["geometry"]))
        else:
            # multipolygon
            for pol in poly_shape:
                for part in pol["geometry"]["coordinates"]:
                    vertices_count += len(part)
                shp_geom.append(shape(pol['geometry']))
        poly_count = len(shp_geom)
        poly_s = ""
        if poly_count > 1:
            poly_s = "s"
        print(
            f'Loaded shapefile with '
            f'{poly_count} polygon{poly_s} '
            f'and '
            f'{vertices_count} vertices.'
        )
        return shp_geom

    def load_custom_crs(self, override_crs):
        """Optionally, create custom crs for projecting
        data base don user input

        Note: not fully implemented
        """
        self.crs_proj = pyproj.Proj(init="epsg:{0}".format(override_crs))
        print("Custom CRS set: " + str(self.crs_proj.srs))
        self.epsg_code = override_crs

    def load_sourcemapping(self):
        """Loads source mapping, if available.

        Otherwise, try to read structure from first line of CSV.
        """
        mapping_config_path = (
            self.config_folder /
            f"sourcemapping_{self.data_source.lower()}.cfg"
        )
        if not os.path.exists(mapping_config_path):
            orig_path = (
                self.config_folder /
                f'sourcemapping_fromlbsn.cfg'
            )
            if os.path.exists(orig_path):
                warnings.warn(
                    "Please rename resources/sourcemapping_fromlbsn.cfg "
                    "to sourcemapping_lbsn.cfg. The former name will be "
                    "deprecated in the future.", DeprecationWarning)
                mapping_config_path = orig_path
            else:
                exit("Source mapping does not exist: "
                     f"sourcemapping_{self.data_source.lower()}.cfg")
        source_config = configparser.ConfigParser()
        source_config.read(mapping_config_path)
        source_config_py = ConfigMap(source_config)
        return source_config_py

    @staticmethod
    def set_glob_options():
        """Includes global options in other packages to be set
        prior execution"""
        # unused


class ConfigMap:
    """Retrieves python object from config.cfg"""

    def __init__(self, source_config):
        # [Main]
        self.name = source_config["Main"]["name"]
        self.file_extension = source_config["Main"]["file_extension"].lower()
        self.delimiter = source_config["Main"]["delimiter"]
        self.array_separator = source_config["Main"]["array_separator"]
        self.quoting = self._quote_selector(
            source_config["Main"]["quoting"])
        self.quote_char = source_config["Main"]["quote_char"].strip('\'')
        self.date_time_format = source_config["Main"]["file_extension"]
        # [Columns]
        self.originid_col = source_config["Columns"]["originid_col"]
        self.post_guid_col = source_config["Columns"]["post_guid_col"]
        self.latitude_col = source_config["Columns"]["latitude_col"]
        self.longitude_col = source_config["Columns"]["longitude_col"]
        self.user_guid_col = source_config["Columns"]["user_guid_col"]
        self.post_create_date_col = \
            source_config["Columns"]["post_create_date_col"]
        self.post_publish_date_col = \
            source_config["Columns"]["post_publish_date_col"]
        self.post_views_count_col = \
            source_config["Columns"]["post_views_count_col"]
        self.post_like_count_col = \
            source_config["Columns"]["post_like_count_col"]
        self.post_url_col = source_config["Columns"]["post_url_col"]
        self.tags_col = source_config["Columns"]["tags_col"]
        self.emoji_col = source_config["Columns"]["emoji_col"]
        self.post_title_col = source_config["Columns"]["post_title_col"]
        self.post_body_col = source_config["Columns"]["post_body_col"]
        self.post_geoaccuracy_col = \
            source_config["Columns"]["post_geoaccuracy_col"]
        self.place_guid_col = source_config["Columns"]["place_guid_col"]
        self.place_name_col = source_config["Columns"]["place_name_col"]

    @staticmethod
    def _quote_selector(quote_string):
        quote_switch = {
            "QUOTE_MINIMAL": csv.QUOTE_MINIMAL,
            "QUOTE_ALL": csv.QUOTE_ALL,
            "QUOTE_NONNUMERIC": csv.QUOTE_NONNUMERIC,
            "QUOTE_NONE": csv.QUOTE_NONE,
        }
        quoting = quote_switch.get(quote_string)
        return quoting



# -*- coding: utf-8 -*-

"""Setup config for cx_freeze (build)

This cx_freeze setup is configured to build
self executable folder on Windows for Windows
users who cannot use package manager. The resulting
tagmaps.exe can be directly run.

Has been tested with conda environment.

Run build with:
    python cx_setup.py build

Extra Requirements for build:
    - install OSGeosW64, add C:/OSGeo4W64/bin to PATH
    - install mkl-service
"""

import glob
import json
import os.path
import opcode
from pathlib import Path
from cx_Freeze import Executable, setup

# Derive Package Paths Dynamically
PYTHON_INSTALL_DIR = os.path.dirname(
    os.path.dirname(os.__file__))
os.environ['TCL_LIBRARY'] = os.path.join(
    PYTHON_INSTALL_DIR, 'tcl', 'tcl8.6')
os.environ['TK_LIBRARY'] = os.path.join(
    PYTHON_INSTALL_DIR, 'tcl', 'tk8.6')

# opcode is not a virtualenv module,
# so we can use it to find the stdlib; this is the same
# trick used by distutils itself it installs itself into the virtualenv
DISTUTILS_PATH = os.path.join(os.path.dirname(opcode.__file__), 'distutils')

VERSION_DICT = {}
with open("tagmaps/version.py") as fp:
    exec(fp.read(), VERSION_DICT)  # pylint: disable=W0122
VERSION = VERSION_DICT['__version__']

# Dependencies are automatically detected,
# but it might need fine tuning.
INCLUDES_MOD = [
    "tkinter.filedialog",
    'numpy.lib.format',
    'numpy.core._methods',
    'matplotlib.backends.backend_tkagg',
    'seaborn',
    'seaborn.cm',
    'scipy.sparse.csgraph',
    'argparse',
    'scipy.sparse.csgraph._validation',
    'numpy',
    'scipy._distributor_init',
    'multiprocessing',
    'gdal',
    'pyproj.datadir',
    'shapely.geometry',
]

# include all mkl files
# https://stackoverflow.com/questions/54337644/cannot-load-mkl-intel-thread-dll-on-python-executable
# https://github.com/anthony-tuininga/cx_Freeze/issues/214
# get json file that has mkl files list (exclude the "service" file)
MKL_FILES_JSON_FILE = glob.glob(
    os.path.join(PYTHON_INSTALL_DIR, "conda-meta", "mkl-[!service]*.json"))[0]
with open(MKL_FILES_JSON_FILE) as file:
    MKL_FILES_JSON_DATA = json.load(file)
# get the list of files from the json data file
NUMPY_MKL_DLLS = MKL_FILES_JSON_DATA["files"]
# get the full path of these files
NUMPY_DLLS_FULLPATH = list(map(lambda currPath: os.path.join(
    PYTHON_INSTALL_DIR, currPath), NUMPY_MKL_DLLS))
# add libiomp5md.dll, as it is also needed for Intel MKL
NUMPY_DLLS_FULLPATH.append(
    os.path.join(PYTHON_INSTALL_DIR, 'Library', 'bin', 'libiomp5md.dll')
)

# files that need manual attention
INCLUDE_FOLDERS_FILES = [
    (DISTUTILS_PATH, 'distutils'),
    (os.path.join(PYTHON_INSTALL_DIR, 'Library', 'bin', 'tcl86t.dll'),
     os.path.join('lib', 'tcl86t.dll'),
     ),
    (os.path.join(PYTHON_INSTALL_DIR, 'Library', 'bin', 'tk86t.dll'),
     os.path.join('lib', 'tk86t.dll'),
     ),
    (os.path.join(PYTHON_INSTALL_DIR, 'Library', 'bin', 'sqlite3.dll'),
     os.path.join('lib', 'sqlite3.dll'),
     ),
    # both files of geos need to be available in both directories
    (os.path.join(PYTHON_INSTALL_DIR, 'Library', 'bin', 'geos.dll'),
     os.path.join('Library', 'lib', 'geos.dll')
     ),
    (os.path.join(PYTHON_INSTALL_DIR, 'Library', 'bin', 'geos_c.dll'),
     os.path.join('Library', 'lib', 'geos_c.dll')
     ),
    (os.path.join(PYTHON_INSTALL_DIR, 'Library', 'bin', 'geos.dll'),
     os.path.join('geos.dll')
     ),
    (os.path.join(PYTHON_INSTALL_DIR, 'Library', 'bin', 'geos_c.dll'),
     os.path.join('geos_c.dll')
     ),
    (os.path.join(PYTHON_INSTALL_DIR, 'Library', 'share', 'proj'),
     os.path.join('proj')
     ),
    os.path.join(os.environ['GDAL_DATA'], 'gcs.csv'),
    'resources/01_Input/',
    'resources/00_Config/',
    'resources/00_generateClusters_OnlyEmoji.cmd',
    'resources/00_generateClusters_OnlyPhotoLocations.cmd',
    'resources/00_generateClusters_OnlyTags.cmd',
    ('tagmaps/matplotlibrc',
     "matplotlibrc")
] + NUMPY_DLLS_FULLPATH

PACKAGES_MOD = ["tkinter", "hdbscan"]
EXCLUDES_MOD = [
    "scipy.spatial.cKDTree",
    "distutils"]


# GUI applications require a different base on Windows
# (the default is for a console application).
BASE = None

EXECUTABLES = [
    Executable('tagmaps/__main__.py',
               base=BASE,
               targetName="tagmaps.exe")
]

BUILD_NAME = f'tagmaps-{VERSION}-win-amd64-3.6'

setup(name="tagmaps",
      version=VERSION,
      description="Tag Clustering for Tag Maps",
      options={
          'build_exe': {
              'includes': INCLUDES_MOD,
              'include_files': INCLUDE_FOLDERS_FILES,
              'packages': PACKAGES_MOD,
              'excludes': EXCLUDES_MOD,
              'optimize': 0,
              'build_exe': (
                  Path.cwd() / 'build' / BUILD_NAME)
          }
      },
      executables=EXECUTABLES)


# open issue Windows:
# after build, rename
# \lib\multiprocessing\Pool.pyc
# to pool.pyc
# see:
# https://github.com/anthony-tuininga/cx_Freeze/issues/353
BUILD_PATH_POOL = (
    Path.cwd() / 'build' / BUILD_NAME /
    'lib' / 'multiprocessing')
Path(BUILD_PATH_POOL / 'Pool.pyc').rename(
    BUILD_PATH_POOL / 'pool.pyc')



# -*- coding: utf-8 -*-

"""Module for matplotlib, seaborn, pyplot methods.
"""

from __future__ import absolute_import

from typing import Tuple

import matplotlib.pyplot as plt
from matplotlib.ticker import FuncFormatter
from descartes import PolygonPatch

from tagmaps.classes.shared_structure import (EMOJI, LOCATIONS, TAGS,
                                              AnalysisBounds)
from tagmaps.classes.utils import Utils


class TPLT():
    """Tag Maps plotting Class

    To remember (because mpl/pyplot can be confusing):
        - The figure is the window that the plot is in.
          It's the top-level container. Each figure has a canvas
          where things are painted on.
        - Each figure usually has one or more axes.
          These are the plots/subplots. Here, only one axes is used.
        - Colorbars and other stuff are also inside the figure.
          Adding a colorbar creates a new axe (unless specified otherwise)
        - two modes exist, OO (object oriented) and
          pyplot("state-machine interface"),
          use OO-Mode if possible because it is more flexible and works
          better with Jupyter Mode
    for more information, see:
    https://stackoverflow.com/questions/19816820/how-to-retrieve-colorbar-instance-from-figure-in-matplotlib
    """
    PLOT_KWDS = {'alpha': 0.5, 's': 10, 'linewidths': 0}

    @staticmethod
    def _get_xy_dists(
            bounds: AnalysisBounds) -> Tuple[float, float]:
        """Get X/Y Distances from Analysis Bounds"""
        dist_y_lat = (
            bounds.lim_lat_max - bounds.lim_lat_min)
        dist_x_lng = (
            bounds.lim_lng_max - bounds.lim_lng_min)
        return dist_y_lat, dist_x_lng

    @staticmethod
    def plt_setxy_lim(axis, bounds: AnalysisBounds):
        """Set global plotting bounds basedon Analysis Bounds"""
        axis.set_xlim(
            [bounds.lim_lng_min, bounds.lim_lng_max])
        axis.set_ylim(
            [bounds.lim_lat_min, bounds.lim_lat_max])

    @staticmethod
    def get_fig_points(fig, points, bounds, point_size=None):
        """Get figure for numpy.ndarray of points"""
        if not fig:
            # create a new figure window
            fig = plt.figure(1)
            fig.add_subplot(111)
        axis = fig.get_axes()[0]
        # only one subplot (nrows, ncols, axnum)
        if point_size is not None:
            axis.scatter(points.T[0], points.T[1],
                         color='red', alpha=0.5, s=point_size, linewidths=0)
        else:
            axis.scatter(points.T[0], points.T[1],
                         color='red', **TPLT.PLOT_KWDS)
        fig.canvas.set_window_title('Preview Map')
        TPLT.plt_setxy_lim(axis, bounds)
        axis.tick_params(labelsize=10)
        return fig

    @staticmethod
    def get_sel_preview(points, item, bounds, cls_type):
        """Returns plt map for item selection preview"""
        # img_ratio = TPLT._get_img_ratio(bounds)
        fig = None
        fig = TPLT.get_fig_points(fig, points, bounds)
        TPLT.set_plt_suptitle(fig, item, cls_type)
        return fig

    @staticmethod
    def get_centroid_preview(points, item, bounds, cls_type, point_size):
        """Returns plt map for item selection preview"""
        # img_ratio = TPLT._get_img_ratio(bounds)
        fig = None
        fig = TPLT.get_fig_points(fig, points, bounds, point_size)
        TPLT.set_plt_suptitle(fig, item, cls_type)
        return fig

    @staticmethod
    def get_cluster_preview(
            points, sel_colors, item_text, bounds, mask_noisy,
            cluster_distance, number_of_clusters, auto_select_clusters=None,
            shapes=None, fig=None, cls_type=None):
        """Get cluster preview figure

        Args:
            points (numpy.ndarray): Point coordinates
            sel_colors ([type]): Cluster colors
            item_text ([type]): Text for suptitle
            bounds ([type]): Bounds of plotting area
            mask_noisy ([type]): Statistics for output of noisy clusters text
            cluster_distance ([type]): Cluster distance
            number_of_clusters (int): Total number of clusters
            auto_select_clusters ([type], optional): Defaults to None.
                If clusters have been auto-selected (no cluster distance)
            shapes ([type], optional): Defaults to None.
                Optional shapes for cluster outline/polygons.
            fig ([type], optional): Defaults to None.
                Optional figure for printing results.
            cls_type ([type], optional): Defaults to None.
                Cluster type, used to format text(s)

        Returns:
           figure [plt.fig]: Mpl Figure object
        """

        if auto_select_clusters is None:
            auto_select_clusters = False
        # img_ratio = TPLT._get_img_ratio(bounds)
        if not fig:
            fig = plt.figure(1)
            fig.add_subplot(111)
        axis = fig.get_axes()[0]
        # create main cluster points map
        axis.scatter(points.T[0], points.T[1],
                     c=sel_colors, **TPLT.PLOT_KWDS)
        fig.canvas.set_window_title('Cluster Preview')
        TPLT.set_plt_suptitle(fig, item_text, cls_type)
        dist_text = ''
        if shapes:
            for shape in shapes:
                patch = TPLT._get_poly_patch(shape)
                axis.add_patch(patch)
        if auto_select_clusters is False:
            dist_text = '@ ' + str(int(cluster_distance)) + 'm'
        axis.set_title(f'Cluster Preview {dist_text}',
                       fontsize=12, loc='center')
        noisy_txt = '{} / {}'.format(mask_noisy.sum(), len(mask_noisy))
        axis.text(bounds.lim_lng_max,
                  bounds.lim_lat_max,
                  f'{number_of_clusters} Clusters (Noise: {noisy_txt})',
                  fontsize=10, horizontalalignment='right',
                  verticalalignment='top', fontweight='bold')
        # set plotting bounds
        TPLT.plt_setxy_lim(axis, bounds)
        TPLT.set_plt_tick_params(axis)
        # define new figure so this one is not
        # overwritten in interactive notebook mode
        # plt.figure()
        return fig

    @staticmethod
    def get_single_linkage_tree_preview(
            item_text, fig, cluster_distance, cls_type):
        """Gets figure for single linkage tree from HDBSCAN results"""
        TPLT.set_plt_suptitle(fig, item_text, cls_type)
        fig.canvas.set_window_title('Single Linkage Tree')
        axis = fig.get_axes()[0]
        axis.set_title('Single Linkage Tree', fontsize=12,
                       loc='center')
        # plot cutting distance
        y_value = Utils.get_radians_from_meters(
            cluster_distance)
        axis.relim()
        xmin = axis.get_xlim()[0]
        xmax = axis.get_xlim()[1]
        axis.plot(
            [xmin, xmax], [y_value, y_value], color='k',
            label=f'Cluster (Cut) Distance {int(cluster_distance)}m'
        )
        axis.legend(fontsize=10)
        # replace y_value labels with meters text (instead of radians dist)
        axis.yaxis.set_major_formatter(FuncFormatter(TPLT.y_formatter))
        TPLT.set_plt_tick_params(axis)
        return fig

    @staticmethod
    def y_formatter(y_value, __):
        """Format radians y-labels as meters for improved legibility"""
        return f'{Utils.get_meters_from_radians(y_value):3.0f}m'

    @staticmethod
    def set_plt_suptitle(fig, item: str, cls_type):
        """Sets suptitle for plot (plt) and Cluster Type"""
        TPLT._set_pltspec_suptitle(
            fig, item, cls_type)

    @staticmethod
    def _set_pltspec_suptitle(fig, item: str, cls_type=None):
        """Sets suptitle for plot (plt)"""
        title = TPLT._get_pltspec_suptitle(item, cls_type)
        if cls_type and cls_type == EMOJI:
            plt.rcParams['font.family'] = 'DejaVu Sans'
        else:
            plt.rcParams['font.family'] = 'sans-serif'
        TPLT._set_plt_suptitle_st(fig, title)

    @staticmethod
    def _set_plt_suptitle_st(fig, title: str):
        """Set title of plt"""
        fig.suptitle(title,
                     fontsize=18, fontweight='bold')

    @staticmethod
    def _get_pltspec_suptitle(item: str, cls_type=None) -> str:
        """Gets formatted suptitle for plot

        - depending on clusterer type
        """
        if cls_type is None:
            cls_type = TAGS
        title = ""
        if cls_type == LOCATIONS:
            title = item.upper()
        elif cls_type == EMOJI:
            emoji_name = Utils.get_emojiname(item)
            title = f'{item} ({emoji_name})'
        else:
            title = item.upper()
        return title

    @staticmethod
    def set_plt_tick_params(axis):
        """Sets common plt tick params"""
        axis.tick_params(labelsize=10)

    @staticmethod
    def _get_poly_patch(polygon):
        """Returns a matplotlib polygon-patch from shapely polygon"""
        patch = PolygonPatch(polygon, fc='#999999',
                             ec='#000000', fill=True,
                             zorder=-1, alpha=0.7)
        return patch

    # @staticmethod
    # def _get_img_ratio(bounds: AnalysisBounds
    #                    ) -> float:
    #     """Gets [img] ratio form bounds."""
    #     dists = TPLT._get_xy_dists(bounds)
    #     dist_y_lat = dists[0]
    #     dist_x_lng = dists[1]
    #     # distYLat = Utils.haversine(limXMin,limYMax,limXMin,limYMin)
    #     # distXLng = Utils.haversine(limXMax,limYMin,limXMin,limYMin)
    #     img_ratio = dist_x_lng/(dist_y_lat*2)
    #     return img_ratio



# -*- coding: utf-8 -*-

"""
Module for tag maps clustering methods
"""

from __future__ import absolute_import

import logging
import sys
import warnings
from collections import defaultdict, namedtuple
from functools import partial, wraps
from multiprocessing.pool import ThreadPool
from typing import Any, Dict, List, NamedTuple, Set, Tuple

import numpy as np
import pandas as pd
import pyproj
import seaborn as sns
import shapely.geometry as geometry
from pyproj import Proj  # pylint: disable=C0412
from shapely.ops import transform  # pylint: disable=C0412

from tagmaps.classes.alpha_shapes import AlphaShapes
from tagmaps.classes.plotting import TPLT
from tagmaps.classes.shared_structure import (EMOJI, LOCATIONS, TAGS, TOPICS,
                                              AnalysisBounds, CleanedPost,
                                              ClusterType)
from tagmaps.classes.utils import Utils

# transformer only available from pyproj>=2.0.0
# load if available to improve projection speed
try:
    from pyproj import Transformer  # pylint: disable=C0412
except ImportError:
    pass

try:
    # filter DeprecationWarning
    # by catching outdated hdbscan imports from sklearn.externals
    sys.modules['sklearn.externals.six'] = __import__('six')
    sys.modules['sklearn.externals.joblib'] = __import__('joblib')
    import hdbscan
except ImportError:
    import hdbscan

POOL = ThreadPool(processes=1)
sns.set_context('poster')
sns.set_style('white')


class ClusterGen():
    """Cluster methods for tags, emoji and post locations

    Note that there are three different projections used here:

    1. Input: Original Data in Decimal Degrees (WGS1984)
    2. Intermediate: Radians data converted from Decimal Degrees with
    np.radians(points) for use in HDBSCAN clustering
    3. Output: Projected coordinates based on auto-selected UTM Zone,
    for calculating Alpha Shapes and writing results to
    shapefile
    """
    class CGDec():
        """Decorators for class CG methods"""
        @staticmethod
        def input_topic_format(func):
            """Check if cluster type is topic and if,
                concat item list to string."""
            @wraps(func)
            def _wrapper(self, item, **kwargs):
                if self.cls_type == TOPICS:
                    if isinstance(item, list):
                        item = Utils.concat_topic(item)
                    elif not '-' in item:
                        raise ValueError(
                            "Please supply either list of terms, or"
                            "concatenate terms with '-' character.")
                return func(self, item, **kwargs)
            return _wrapper

    def __init__(self, bounds: AnalysisBounds,
                 cleaned_post_dict: Dict[str, CleanedPost],
                 cleaned_post_list: List[CleanedPost],
                 top_list: List[Tuple[str, int]],
                 total_distinct_locations: int,
                 cluster_type: ClusterType = TAGS,
                 local_saturation_check: bool = False):
        self.cls_type = cluster_type
        self.bounds = bounds
        self.cluster_distance = ClusterGen._init_cluster_dist(
            self.bounds, self.cls_type)
        self.cleaned_post_dict = cleaned_post_dict
        self.cleaned_post_list = cleaned_post_list
        self.top_list = top_list
        if self.top_list:
            self.top_item = top_list[0]
        else:
            self.top_item = None
        self.total_distinct_locations = total_distinct_locations
        self.autoselect_clusters = False  # no cluster distance needed
        self.clusterer = None
        self.local_saturation_check = local_saturation_check
        # storing cluster results:
        self.single_items_dict = defaultdict(list)
        self.clustered_items_dict = defaultdict(list)
        self.clustered_guids_all: List[str] = list()
        self.none_clustered_guids: List[str] = list()
        # get initial analysis bounds in Decimal Degrees
        # for calculating output UTM Zone Projection
        self._update_bounds()
        self.bound_points_shapely = Utils.get_shapely_bounds(
            self.bounds)
        # data always in lat/lng WGS1984
        Utils.set_proj_dir()
        # define input and UTM projections
        self.crs_wgs = Proj(init='epsg:4326', preserve_units=False)
        self.crs_proj, __ = Utils.get_best_utmzone(
            self.bound_points_shapely)
        # define projection function ahead, if available
        if 'pyproj.transformer' in sys.modules:
            self.proj_transformer = Transformer.from_proj(
                self.crs_wgs, self.crs_proj)
            self.proj_transformer_back = Transformer.from_proj(
                self.crs_proj, self.crs_wgs)
        else:
            self.proj_transformer = None
            self.proj_transformer_back = None
        # partial used for geometry projection
        self.proj_transformer_partial = partial(
            pyproj.transform,
            self.crs_wgs,
            self.crs_proj)
        self.proj_transformer_partial_back = partial(
            pyproj.transform,
            self.crs_proj,
            self.crs_wgs)

    @classmethod
    def new_clusterer(cls,
                      cls_type: ClusterType,
                      bounds: AnalysisBounds,
                      cleaned_post_dict: Dict[str, CleanedPost],
                      cleaned_post_list: List[CleanedPost],
                      cleaned_stats: NamedTuple,
                      local_saturation_check: bool):
        """Create new clusterer from type and input data

        Args:
            cls_type (ClusterType): Either TAGS,
                LOCATIONS, TOPICS or EMOJI
            bounds (LoadData.AnalysisBounds): Analaysis spatial boundary
            cleaned_post_dict (Dict[str, CleanedPost]): Dict of cleaned posts
            prepared_data (LoadData.PreparedData): Statistics data

        Returns:
            clusterer (ClusterGen): A new clusterer of ClusterType
        """
        clusterer = cls(
            bounds=bounds,
            cleaned_post_dict=cleaned_post_dict,
            cleaned_post_list=cleaned_post_list,
            top_list=cleaned_stats[cls_type].top_items_list,
            total_distinct_locations=cleaned_stats[
                LOCATIONS].total_unique_items,
            cluster_type=cls_type,
            local_saturation_check=local_saturation_check)
        return clusterer

    @staticmethod
    def _init_cluster_dist(bounds: AnalysisBounds,
                           cls_type: ClusterType) -> float:
        """Get initial cluster distance from analysis bounds.

        - 7% of research area width/height (max) = optimal
        - default value #223.245922725 #= 0.000035 radians dist
        """
        dist_y = Utils.haversine(bounds.lim_lng_min,
                                 bounds.lim_lat_min,
                                 bounds.lim_lng_min,
                                 bounds.lim_lat_max)
        dist_x = Utils.haversine(bounds.lim_lng_min,
                                 bounds.lim_lat_min,
                                 bounds.lim_lng_max,
                                 bounds.lim_lat_min)
        cluster_distance = (min(dist_x, dist_y)/100)*7
        if cls_type == LOCATIONS:
            # since location clustering includes
            # all data, use reduced default distance
            cluster_distance = cluster_distance/8
        return cluster_distance

    def _update_bounds(self):
        """Update analysis rectangle boundary based on

        cleaned posts list."""
        dataframe = pd.DataFrame(self.cleaned_post_list)
        # get columns lng, lat
        # convert to numpy ndarray
        # (List of [lng, lat] lists)
        points = dataframe.loc[:, ['lng', 'lat']].to_numpy()
        (self.bounds.lim_lat_min,
         self.bounds.lim_lat_max,
         self.bounds.lim_lng_min,
         self.bounds.lim_lng_max) = Utils.get_rectangle_bounds(points)

    def _select_postguids(self, item: str) -> Tuple[List[str], int]:
        """Select all posts that have a specific item

        Args:
            item: tag, emoji, location

        Returns:
            Tuple[List[str], int]: list of post_guids and
                                   number of distinct locations
        """
        SelectedItems = namedtuple(
            'SelectedItems',
            'guids location_count')
        distinct_localloc_count = set()
        selected_postguids_list = list()
        for cleaned_post_location in self.cleaned_post_list:
            if self.cls_type == TAGS:
                self._filter_tags(
                    item, cleaned_post_location,
                    selected_postguids_list,
                    distinct_localloc_count)
            elif self.cls_type == EMOJI:
                self._filter_emoji(
                    item, cleaned_post_location,
                    selected_postguids_list,
                    distinct_localloc_count)
            elif self.cls_type == LOCATIONS:
                self._filter_locations(
                    item, cleaned_post_location,
                    selected_postguids_list,
                    distinct_localloc_count)
            elif self.cls_type == TOPICS:
                self._filter_topics(
                    item, cleaned_post_location,
                    selected_postguids_list,
                    distinct_localloc_count)
            else:
                raise ValueError(f"Clusterer {self.cls_type} unknown.")
        selected_items = SelectedItems(
            selected_postguids_list, len(distinct_localloc_count))
        return selected_items

    @staticmethod
    def _filter_tags(
            item: str,
            cleaned_photo_location: CleanedPost,
            selected_postguids_list: List[str],
            distinct_localloc_count: Set[str]):
        if (item in (cleaned_photo_location.hashtags) or
                (item in cleaned_photo_location.post_body)):
            selected_postguids_list.append(
                cleaned_photo_location.guid)
            distinct_localloc_count.add(
                cleaned_photo_location.loc_id)

    @staticmethod
    def _filter_topics(
            item: str,
            cleaned_photo_location: CleanedPost,
            selected_postguids_list: List[str],
            distinct_localloc_count: Set[str]):
        """Check topics against tags, body and emoji"""
        item_list = Utils.split_topic(item)
        if (ClusterGen._compare_anyinlist(
                item_list, cleaned_photo_location.hashtags)
                or ClusterGen._compare_anyinlist(
                    item_list, cleaned_photo_location.post_body)
                or ClusterGen._compare_anyinlist(
                    item_list, cleaned_photo_location.emoji)):
            selected_postguids_list.append(
                cleaned_photo_location.guid)
            distinct_localloc_count.add(
                cleaned_photo_location.loc_id)

    @staticmethod
    def _compare_anyinlist(items, item_list):
        """Check if any term of topic is in list"""
        if any(x in items for x in item_list):
            return True
        return False

    @staticmethod
    def _filter_emoji(
            item: str,
            cleaned_photo_location: CleanedPost,
            selected_postguids_list: List[str],
            distinct_localloc_count: Set[str]):
        if item in cleaned_photo_location.emoji:
            selected_postguids_list.append(
                cleaned_photo_location.guid)
            distinct_localloc_count.add(
                cleaned_photo_location.loc_id)

    @staticmethod
    def _filter_locations(
            item: str,
            cleaned_photo_location: CleanedPost,
            selected_postguids_list: List[str],
            distinct_localloc_count: Set[str]):
        if item == cleaned_photo_location.loc_id:
            selected_postguids_list.append(
                cleaned_photo_location.guid)
            distinct_localloc_count.add(
                cleaned_photo_location.loc_id)

    def _getselect_postguids(self, item: str,
                             silent: bool = True) -> List[str]:
        """Get list of post guids with specific item

        Args:
            item: tag, emoji, location
        """
        sel_items = self._select_postguids(item)
        if silent:
            return sel_items.guids
        # console reporting
        if self.cls_type == EMOJI:
            item_text = Utils.get_emojiname(item)
        else:
            item_text = item
        type_text = self.cls_type.rstrip('s')
        perc_oftotal_locations = (
            sel_items.location_count /
            (self.total_distinct_locations/100)
        )
        perc_text = ""
        if perc_oftotal_locations >= 1:
            perc_text = (f'(found in {perc_oftotal_locations:.0f}% '
                         f'of DLC in area)')
        item_index_pos = self._get_toplist_index(item) + 1
        print(f"({item_index_pos} of {len(self.top_list)}) "
              f"Found {len(sel_items.guids)} posts (UPL) "
              f"for {type_text} '{item_text}' "
              f"{perc_text}", end=" ")
        return sel_items.guids

    def _get_toplist_index(self, item_text: str) -> int:
        """Get Position of Item in Toplist"""
        try:
            index_pos = Utils.get_index_of_tup(
                self.top_list, 0, item_text)
        except ValueError:
            index_pos = 0
        return index_pos

    def _getselect_posts(self,
                         selected_postguids_list: List[str]
                         ) -> List[CleanedPost]:
        selected_posts_list = [self.cleaned_post_dict[x]
                               for x in selected_postguids_list]
        return selected_posts_list

    def get_np_points_guids(self, item: str = None,
                            silent: bool = None, sel_all: bool = None
                            ) -> np.ndarray:
        """Gets numpy array of selected points with latlng containing _item

        Args:
            item: tag, emoji, location; or topic (list of terms)
            silent: if true, no console output (interface mode)

        Returns:
            points: A list of lat/lng points to map
            selected_postguids_list: List of selected post guids
        """
        SelItems = namedtuple(
            'SelItems',
            'points guids')
        # no log reporting for selected points
        if silent is None:
            silent = False
        if sel_all is None:
            sel_all = False
        if sel_all:
            # select all post guids
            selected_postguids_list = list()
            for cleaned_post in self.cleaned_post_list:
                selected_postguids_list.append(
                    cleaned_post.guid)
            selected_posts_list = self.cleaned_post_list
        else:
            selected_postguids_list = self._getselect_postguids(
                item, silent=silent)
            # clustering
            if len(selected_postguids_list) < 2:
                # return empty list of points
                return SelItems([], selected_postguids_list)
            selected_posts_list = self._getselect_posts(
                selected_postguids_list)
        # only used for tag clustering,
        # otherwise (photo location clusters),
        # global vars are used (dataframe, points)
        dataframe = pd.DataFrame(selected_posts_list)
        # converts pandas data to numpy array
        # (limit by list of column-names)
        points = dataframe.loc[:, ['lng', 'lat']].to_numpy()
        # only return preview fig without clustering
        return SelItems(points, selected_postguids_list)

    def get_np_points(self, item: str = None, silent: bool = None
                      ) -> np.ndarray:
        """Wrapper that only returns points for _get_np_points_guids"""
        # decide if select all or specific item
        sel_all = bool(item is None)
        sel_items = self.get_np_points_guids(item, silent, sel_all)
        if sel_items.points.size:
            return sel_items.points

    def _cluster_points(self, points,
                        min_span_tree: bool = None,
                        preview_mode: bool = None,
                        min_cluster_size: int = None,
                        allow_single_cluster: bool = True):
        """Cluster points using HDBSCAN"""
        if min_span_tree is None:
            min_span_tree = False
        if preview_mode is None:
            preview_mode = False
        if allow_single_cluster is None:
            allow_single_cluster = True
        # conversion to radians for HDBSCAN
        tag_radians_data = np.radians(points)  # pylint: disable=E1111
        if min_cluster_size is None:
            min_cluster_size = max(
                2, int(((len(points))/100)*5))
        self.clusterer = hdbscan.HDBSCAN(
            min_cluster_size=min_cluster_size,
            gen_min_span_tree=min_span_tree,
            allow_single_cluster=allow_single_cluster,
            min_samples=1)
        # Start clusterer on different thread
        # to prevent GUI from freezing
        with warnings.catch_warnings():
            # disable joblist multithread warning
            # because there's only one thread
            warnings.simplefilter('ignore', UserWarning)
            async_result = POOL.apply_async(
                ClusterGen._fit_cluster, (self.clusterer, tag_radians_data))
            self.clusterer = async_result.get()

        if self.autoselect_clusters:
            cluster_labels = self.clusterer.labels_
        else:
            cluster_labels = self.clusterer.single_linkage_tree_.get_clusters(
                Utils.get_radians_from_meters(
                    self.cluster_distance), min_cluster_size=2)
        # exit function in case of
        # final processing loop (no figure generating)
        if not preview_mode:
            return cluster_labels, None, None, None
        # verbose reporting if preview mode
        mask_noisy = (cluster_labels == -1)
        number_of_clusters = len(
            np.unique(cluster_labels[~mask_noisy]))  # nopep8 false positive? pylint: disable=E1130
        palette = sns.color_palette("husl", number_of_clusters+1)
        sel_colors = [palette[x] if x >= 0
                      else (0.5, 0.5, 0.5)
                      # for x in clusterer.labels_ ]
                      for x in cluster_labels]
        # return additional information in preview mode
        # for plotting
        return cluster_labels, sel_colors, mask_noisy, number_of_clusters

    def cluster_item(self, item: str, preview_mode=None):
        """Cluster specific item

        Args:
            item (str): The item to select and cluster
            preview_mode ([type], optional): Defaults to None. If True,
                sel_colors, mask_noisy, number_of_clusters will be returned,
                which can be used as additional information during plot

        Returns:
            clusters: The cluster labels returned from HDBSCAN
            selected_post_guids: All selected post guids for item
            points: numpy.ndarray of selected post coordinates (radians)
            sel_colors: color codes assigned to points for plotting clusters
            mask_noisy: number of clusters that were ambiguous (from HDBSCAN)
            number_of_clusters: number of identified clusters (from HDBSCAN)
        """
        ClusterResults = namedtuple(
            'ClusterResults',
            'labels guids points colors mask_noisy cluster_count')
        if preview_mode is None:
            preview_mode = False
        sel_items = self.get_np_points_guids(
            item=item, silent=preview_mode)

        if len(sel_items.guids) < 2:
            # no need to cluster
            return None
        (clusters, sel_colors,
         mask_noisy, number_of_clusters) = self._cluster_points(
             points=sel_items.points, preview_mode=preview_mode)
        return ClusterResults(
            clusters, sel_items.guids,
            sel_items.points, sel_colors, mask_noisy, number_of_clusters)

    def _cluster_all_items(self):
        """Cluster all items (e.g. all locations)"""
        ClusterResults = namedtuple(
            'ClusterResults',
            'labels guids')
        sel_items = self.get_np_points_guids(
            silent=False, sel_all=True)
        # min_cluster_size = 2 (LOCATIONS)
        # do not allow clusters with one item
        if len(sel_items.guids) < 2:
            return
        cluster_labels, _, _, _ = self._cluster_points(
            points=sel_items.points, preview_mode=False,
            min_cluster_size=2, allow_single_cluster=False)
        return ClusterResults(cluster_labels, sel_items.guids)

    @staticmethod
    def _get_cluster_guids(clusters, selected_post_guids):
        """Returns two lists: clustered and non clustered guids"""
        Guids = namedtuple(
            'Guids',
            'clustered nonclustered')
        clustered_guids = list()
        np_selected_post_guids = np.asarray(selected_post_guids)
        mask_noisy = (clusters == -1)
        if len(selected_post_guids) == 1:
            number_of_clusters = 0
        else:
            number_of_clusters = len(np.unique(clusters[~mask_noisy]))
        if number_of_clusters == 0:
            print("--> No cluster.")
            none_clustered_guids = list(np_selected_post_guids)
        else:
            print(f'--> {number_of_clusters} cluster.')
            for cluster_x in range(number_of_clusters):
                current_clustered_guids = np_selected_post_guids[clusters == cluster_x]
                clustered_guids.append(current_clustered_guids)
            none_clustered_guids = list(np_selected_post_guids[clusters == -1])
            # Sort descending based on size of cluster
            # this is needed to later compute HImp Value (1 or 0)
            clustered_guids.sort(key=len, reverse=True)
        return Guids(clustered_guids, none_clustered_guids)

    def _get_update_clusters(self, item: str = None,
                             single_items_dict=None,
                             cluster_items_dict=None,
                             itemized: bool = None):
        """Get clusters for items and write results to dicts"""
        if not single_items_dict:
            single_items_dict = self.single_items_dict
        if not cluster_items_dict:
            cluster_items_dict = self.clustered_items_dict
        if itemized is None:
            # default
            itemized = True
        if itemized:
            # clusters guids points colors mask_noisy cluster_count
            cluster = self.cluster_item(item)
        else:
            cluster = self._cluster_all_items()
        if not cluster:
            print("--> No cluster (all locations removed).")
            return
        # get clustered guids/ non-clustered guids
        guids = self._get_cluster_guids(cluster.labels, cluster.guids)
        if itemized:
            single_items_dict[item] = guids.nonclustered
            if guids.clustered:
                cluster_items_dict[item] = guids.clustered
            # dicts modified in place, no need to return
            return
        else:
            self.clustered_guids_all = guids.clustered
            self.none_clustered_guids = guids.nonclustered

    def get_overall_clusters(self):
        """Get clusters for all items attached to self

        Updates results as two lists:
            self.clustered_guids_all
            self.none_clustered_guids
        """
        # update in case of locations removed
        # self.cleaned_post_list = list(
        #     self.cleaned_post_dict.values())
        self._get_update_clusters(itemized=False)

    def get_itemized_clusters(self):
        """Get itemized clusters for top_list attached to self

        Updates results as two Dict of Lists:
            self.single_items_dict
            self.clustered_items_dict
        """
        # get clusters for top item
        if self.local_saturation_check:
            self._get_update_clusters(
                item=self.top_item)
        tnum = 0
        # get remaining clusters
        for item in self.top_list:
            if (self.local_saturation_check and
                    tnum == 0):
                # skip topitem if already
                # clustered due to local saturation
                continue
            tnum += 1
            self._get_update_clusters(
                item=item.name)
        # logging.getLogger("tagmaps").info(
        #    f'{len(self.clustered_items)} '
        #    f'{self.cls_type.rstrip("s")} clusters.\n'
        #    f'{len(self.single_items)} without neighbors.')
        # flush console output once
        sys.stdout.flush()

    def get_all_cluster_centroids(self) -> NamedTuple:
        """Get all centroids for clustered data

        Returns:
            NamedTuple: Results as named tuple
                        data: shapes and meta information
                        cls_type: ClusterGen [EMOJI, TAGS etc.)
                        itemized: bool
        """

        itemized = False
        ClusterShapes = namedtuple(
            'ClusterShapes',
            'data cls_type itemized')
        cluster_guids = self.clustered_guids_all
        none_clustered_guids = self.none_clustered_guids
        resultshapes_and_meta = self.get_cluster_centroids(
            cluster_guids, none_clustered_guids)
        return ClusterShapes(resultshapes_and_meta, self.cls_type, itemized)

    def get_item_cluster_centroids(self, item, single_clusters=None):
        """Get centroids for item clustered data"""
        if single_clusters is None:
            single_clusters = True
        self._get_update_clusters(
            item=item)
        cluster_guids = self.clustered_items_dict[item]
        if single_clusters:
            none_clustered_guids = self.single_items_dict[item]
        else:
            none_clustered_guids = None
        resultshapes_and_meta = self.get_cluster_centroids(
            cluster_guids, none_clustered_guids)
        return resultshapes_and_meta

    def _proj_coords(self, lng: float, lat: float):
        """Project coordinates based on available packages

        If pyproj > 2.0.0, pyproj.transformer is available,
        which provides a more convenient and faster way to
        project many coordinates.

        However, pyproj <2.0.0 does not support pyproj.transformer,
        in this case (self.proj_transformer is None) use pyproj.transform
        directly.
        """
        if self.proj_transformer:
            lng_proj, lat_proj = self.proj_transformer.transform(
                lng, lat)
        else:
            lng_proj, lat_proj = pyproj.transform(self.crs_wgs, self.crs_proj,
                                                  lng, lat)
        return lng_proj, lat_proj

    def get_cluster_centroids(
            self, clustered_guids, none_clustered_guids=None):
        """Get centroids for clustered data"""
        resultshapes_and_meta = list()
        for post_cluster in clustered_guids:
            posts = [self.cleaned_post_dict[x] for x in post_cluster]
            unique_user_count = len(set([post.user_guid for post in posts]))
            # get points and project coordinates to suitable UTM
            points = [geometry.Point(
                self._proj_coords(post.lng, post.lat)
            ) for post in posts]
            point_collection = geometry.MultiPoint(list(points))
            # convex hull enough for calculating centroid
            result_polygon = point_collection.convex_hull
            result_centroid = result_polygon.centroid
            if result_centroid is not None and not result_centroid.is_empty:
                resultshapes_and_meta.append(
                    (result_centroid, unique_user_count)
                )
        if not none_clustered_guids:
            return resultshapes_and_meta
        # noclusterphotos = [cleanedPhotoDict[x] for x in singlePhotoGuidList]
        for no_cluster_post in none_clustered_guids:
            post = self.cleaned_post_dict[no_cluster_post]
            x_point, y_point = self._proj_coords(
                post.lng, post.lat)
            p_center = geometry.Point(x_point, y_point)
            if p_center is not None and not p_center.is_empty:
                resultshapes_and_meta.append((p_center, 1))
        sys.stdout.flush()
        # log.debug(f'{resultshapes_and_meta[:10]}')
        return resultshapes_and_meta

    def _get_item_clustershapes(
            self,
            item: Tuple[str, int],
            cluster_guids=None) -> Tuple[List[Tuple[Any]], float]:
        """Get Cluster Shapes from a list of coordinates
        for a given item"""
        if cluster_guids is None:
            cluster_guids = self.clustered_items_dict.get(
                item.name, None)
        if not cluster_guids:
            return None, 0
        alphashapes_data = AlphaShapes.get_cluster_shape(
            item, cluster_guids, self.cleaned_post_dict,
            self.cluster_distance,
            self.local_saturation_check, self._proj_coords)
        return alphashapes_data

    def _get_item_clusterarea(
            self,
            item: Tuple[str, int]) -> float:
        """Wrapper: only get cluster shape area for item"""
        alphashape_data = self._get_item_clustershapes(item)
        return alphashape_data.item_area

    @staticmethod
    def _is_saturated_item(
            item_area: float,
            topitem_area: float):
        """Skip item entirely if saturated, i.e.
        if total area > 80%
        of top item cluster area

        Args:
            item_area: item cluster area
            topitem_area: top item cluster area
        """
        local_saturation = item_area/(topitem_area/100)
        # print("Local Saturation for Tag " + self.top_item "
        #       "+ ": " + str(round(localSaturation,0)))
        if local_saturation > 60:
            return True
        else:
            return False

    def _get_item_shapeslist(self, item, topitem_area, tnum):
        """Get all item shapes for item clusters

        Note: A function ref to self._proj_coords is handed
        to AlphaShapes.get_single_cluster_shape(). Coordinates are then
        projected inside AlphaShapes Class, depending on the pyproj version.
        """
        resultshapes_and_meta_tmp = list()
        result = self._get_item_clustershapes(item)
        shapes_tmp = result[0]
        item_area = result[1]
        if (self.local_saturation_check
                and item_area != 0
                and tnum != 1):
            if self._is_saturated_item(item_area,
                                       topitem_area):
                # next item
                return None
        # append result
        if shapes_tmp and shapes_tmp:
            resultshapes_and_meta_tmp.extend(
                shapes_tmp)
        # get shapes for single items (non-clustered)
        none_clustered_guids = self.single_items_dict.get(
            item.name, None)
        if not none_clustered_guids:
            return resultshapes_and_meta_tmp
        posts = [self.cleaned_post_dict[x]
                 for x in none_clustered_guids]
        for single_post in posts:
            shapes_single_tmp = AlphaShapes.get_single_cluster_shape(
                item, single_post, self.cluster_distance,
                self._proj_coords)
            if not shapes_single_tmp:
                continue
            # Use append, since always single Tuple
            resultshapes_and_meta_tmp.append(
                shapes_single_tmp)
        return resultshapes_and_meta_tmp

    def get_cluster_shapes(self):
        """For each cluster of points,
        calculate boundary shape and
        add statistics (HImpTag etc.)

        Returns results as shapes_and_meta
        = list(), ClusterType, itemized = bool
        """
        itemized = True
        saturation_exclude_count = 0
        shapes_and_meta = list()
        tnum = 0
        topitem_area = None
        if self.local_saturation_check:
            # calculate total area of Top1-Tag
            # for 80% saturation check for lower level tags
            topitem_area = self._get_item_clusterarea(
                self.top_item)
            if topitem_area == 0:
                raise ValueError(
                    f'Something went wrong: '
                    f'Could not get area for Top item '
                    f'{self.top_item}')
        for item in self.top_list:
            tnum += 1
            shapes_tmp = self._get_item_shapeslist(
                item, topitem_area, tnum)
            if shapes_tmp is None:
                saturation_exclude_count += 1
                continue
            if not shapes_tmp:
                continue
            shapes_and_meta.extend(shapes_tmp)
        logging.getLogger("tagmaps").info(
            f'{len(shapes_and_meta)} '
            f'alpha shapes. Done.')
        if saturation_exclude_count > 0:
            logging.getLogger("tagmaps").info(
                f'Excluded {saturation_exclude_count} '
                f'{self.cls_type.rstrip("s")} on local saturation check.')
        return shapes_and_meta, self.cls_type, itemized

    @staticmethod
    def _fit_cluster(clusterer, data):
        """Perform HDBSCAN clustering from features or distance matrix.

        Args:
            clusterer ([type]): HDBScan clusterer
            data ([type]): A feature array (points)

        Returns:
            [type]: Clusterer
        """

        clusterer.fit(data)
        return clusterer

    @CGDec.input_topic_format
    def get_sel_preview(self, item):
        """Returns plt map for item selection preview"""
        points = self.get_np_points(
            item=item,
            silent=True)
        fig = TPLT.get_sel_preview(
            points, item, self.bounds, self.cls_type)
        return fig

    @CGDec.input_topic_format
    def get_cluster_centroid_data(
            self, item, zipped=None, projected=None, single_clusters=None):
        """Returns centroids for cluster selection based on item

        Args:
            item (str or list of str): Item to be selected
            zipped ([type], optional): Will merge centroids and user_count,
                                       defaults to False
            projected (bool, optional): Will return projected data (UTM),
                                        otherwise, centroids are returned
                                        in decimal degrees (WGS1984),
                                        defaults to False
            single_clusters: Return single item cluster centroids,
                                        defaults to True

        Returns:
            Tuple: [0] point (List of coordinate pairs),
                   [1] user_count (count of user_count per centroid)
        """
        if zipped is None:
            zipped = False
        if projected is None:
            projected = False
        if single_clusters is None:
            single_clusters = True
        shapes = self.get_item_cluster_centroids(
            item=item, single_clusters=single_clusters)
        points = [meta[0] for meta in shapes]
        user_count = [meta[1] for meta in shapes]
        if not projected:
            # AlphaShapes automatically projects data
            # to compute shapes. If no projection is
            # requested, we have to convert it back to
            # original WGS1984 decimal degrees data
            points = self._project_centroids_back(points)
        # extract centroid coordinates from
        # shapely geometry.Point
        latlng_list = [[point.x, point.y] for point in points]
        # convert coords to numpy.nd array
        points = np.array(latlng_list)
        if zipped:
            zip_list = []
            zip_list = list()
            x_id = 0
            for point in points:
                zip_list.append((point[0], point[1], user_count[x_id]))
                x_id += 1
            result = np.asarray(zip_list)
        else:
            result = (points, user_count)
        return result

    @CGDec.input_topic_format
    def get_cluster_centroid_preview(self, item, single_clusters=None):
        """Returns plt map for item selection cluster centroids"""
        if single_clusters is None:
            single_clusters = True
        points, user_count = self.get_cluster_centroid_data(
            item=item, single_clusters=single_clusters)
        fig = TPLT.get_centroid_preview(
            points, item, self.bounds, self.cls_type, user_count)
        return fig

    @CGDec.input_topic_format
    def get_cluster_preview(self, item):
        """Returns plt map for item cluster preview"""
        points = self.get_np_points(
            item=item,
            silent=True)
        self._cluster_points(
            points=points,
            preview_mode=True)
        (_, _, points, sel_colors,
         mask_noisy, number_of_clusters) = self.cluster_item(
             item=item,
             preview_mode=True)
        fig = TPLT.get_cluster_preview(
            points=points, sel_colors=sel_colors, item_text=item,
            bounds=self.bounds, mask_noisy=mask_noisy,
            cluster_distance=self.cluster_distance,
            number_of_clusters=number_of_clusters,
            auto_select_clusters=self.autoselect_clusters,
            cls_type=self.cls_type)
        return fig

    @CGDec.input_topic_format
    def get_clustershapes_preview(self, item):
        """Returns plt map for item cluster preview"""
        # selected post guids: all posts for item
        # points: numpy-points for plotting
        # clusters: hdbscan labels for clustered items
        result = self.cluster_item(
            item=item,
            preview_mode=True)
        if not result:
            return print("No items found.")
        clusters = result[0]
        selected_post_guids = result[1]
        points = result[2]
        sel_colors = result[3]
        mask_noisy = result[4]
        number_of_clusters = result[5]
        # cluster_guids: those guids that are clustered
        cluster_guids, _ = self._get_cluster_guids(
            clusters, selected_post_guids)
        shapes, _ = self._get_item_clustershapes(item, cluster_guids)
        # get only shapely shapes, not usercount and other info
        shapes = [meta[0] for meta in shapes]
        shapes_wgs = self._project_centroids_back(shapes)
        fig = TPLT.get_cluster_preview(
            points=points, sel_colors=sel_colors, item_text=item,
            bounds=self.bounds, mask_noisy=mask_noisy,
            cluster_distance=self.cluster_distance,
            number_of_clusters=number_of_clusters,
            auto_select_clusters=self.autoselect_clusters,
            shapes=shapes_wgs, cls_type=self.cls_type)
        return fig

    def _project_centroids_back(self, shapes):
        """Proj shapes back to WGS1984 for plotting in matplotlib

        simple list comprehension with projection:
        """
        project = self.proj_transformer_partial_back
        shapes_wgs = [(ClusterGen._project_geometry(
            shape, project)) for shape in shapes]
        return shapes_wgs

    @staticmethod
    def _project_geometry(geom_shape, project):
        # geom_shape_proj = project.transform(geom_shape)
        geom_shape_proj = transform(project, geom_shape)
        return geom_shape_proj

    def get_singlelinkagetree_preview(self, item):
        """Returns figure for single linkage tree from HDBSCAN clustering"""
        if self.cls_type == TOPICS:
            item = Utils.concat_topic(item)
        (_, _, _, _, _, number_of_clusters) = self.cluster_item(
            item=item,
            preview_mode=True)
        axis = self.clusterer.single_linkage_tree_.plot(
            truncate_mode='lastp',
            p=max(50, min(number_of_clusters*10, 256)))
        fig = TPLT.get_single_linkage_tree_preview(
            item, axis.figure, self.cluster_distance,
            self.cls_type)
        return fig



# -*- coding: utf-8 -*-

"""
Setup for building and installing tagmaps package.
"""

import sys
from setuptools import setup

with open('README.md') as f:
    LONG_DESCRIPTION = f.read()

try:
    from semantic_release import setup_hook
    setup_hook(sys.argv)
except ImportError:
    pass

VERSION_NUMBER = {}
with open("tagmaps/version.py") as fp:
    exec(fp.read(), VERSION_NUMBER)  # pylint: disable=W0122

setup(name="tagmaps",
      version=VERSION_NUMBER['__version__'],
      description="Tag Clustering for Tag Maps",
      long_description=LONG_DESCRIPTION,
      long_description_content_type='text/markdown',
      author='Alexander Dunkel',
      author_email='alexander.dunkel@tu-dresden.de',
      url='https://github.com/Sieboldianus/TagMaps',
      license='GNU GPLv3 or any higher',
      packages=['tagmaps'],
      include_package_data=True,
      install_requires=[
          'fiona',
          'shapely',
          'pandas>=0.24.2',
          'pyproj>=1.9.6',
          'numpy',
          'matplotlib',
          'emoji>=0.5.1',
          'cython',
          'hdbscan>=0.8.20',
          'gdal',
          'attrs',
          'seaborn',
          'descartes',
          'regex',
          'scipy',
          'six',
          'joblib',
      ],
      entry_points={
          'console_scripts': [
              'tagmaps = tagmaps.__main__:main'
          ]
      })



# -*- coding: utf-8 -*-

"""
Module for compiling TagMaps results and writing output
"""

from __future__ import absolute_import

from operator import itemgetter
from pathlib import Path
from typing import Dict, List, TextIO, Tuple
from math import sqrt

import shapely.geometry as geometry
import fiona
from fiona.crs import from_epsg

from tagmaps.classes.shared_structure import (EMOJI,
                                              AnalysisBounds)
from tagmaps.classes.utils import Utils


class Compile():
    """Compile results into shapefiles, add statistics, normalize."""
    @classmethod
    def write_shapes(cls,
                     bounds: AnalysisBounds,
                     shapes_and_meta_list,
                     output_folder: Path):
        """Main wrapper for writing
        all results to output

        shapes_and_meta_list is either:
        - List[[Tuple[List[],cls_type, itemized: bool = True]]]
        or:
          List[[Tuple[List[],cls_type, itemized: bool = False]]]
          (overall clusters)
        - List[] contains clustered shapes from ClusterGen and
            attached statistic information
        TODO: refactor into compile & write shapes; update output_folder
        """
        if output_folder:
            print("")
        bound_points_shapely = Utils.get_shapely_bounds(
            bounds)
        # data always in lat/lng WGS1984
        __, epsg_code = Utils.get_best_utmzone(
            bound_points_shapely)
        cls._compile_merge_shapes(
            shapes_and_meta_list, epsg_code, output_folder)

    @classmethod
    def _compile_merge_shapes(cls, shapes_and_meta_list,
                              epsg_code, output_folder):
        all_itemized_shapes = list()
        all_non_itemized_shapes = list()
        contains_emoji_output = False
        for shapes, cls_type, itemized in shapes_and_meta_list:
            if not shapes:
                continue
            if itemized:
                if cls_type == EMOJI:
                    contains_emoji_output = True
                # normalize types separately (e.g. emoji/tags)
                global_weights = cls._get_weights(shapes, [6, 7, 8])
                itemized_shapes = cls._getcompile_itemized_shapes(
                    shapes, cls_type, global_weights)
                # print(f'type itemized_shapes: {type(itemized_shapes)}\n')
                all_itemized_shapes.extend(itemized_shapes)
            else:
                global_weights = cls._get_weights(shapes, [1])
                non_itemized_shapes = cls._getcompile_nonitemized_shapes(
                    shapes, global_weights)
                all_non_itemized_shapes.extend(non_itemized_shapes)
        # writing step:
        if all_itemized_shapes:
            cls._write_all(
                all_itemized_shapes, True,
                contains_emoji_output, epsg_code, output_folder)
        if all_non_itemized_shapes:
            cls._write_all(
                all_non_itemized_shapes, False,
                contains_emoji_output, epsg_code, output_folder)

    @staticmethod
    def _get_shape_schema(itemized):
        """Define polygon feature geometry"""
        if itemized:
            schema = {
                'geometry': 'Polygon',
                'properties': {'Join_Count': 'int',
                               'Views': 'int',
                               'COUNT_User': 'int',
                               'ImpTag': 'str',
                               'TagCountG': 'int',
                               'HImpTag': 'int',
                               'Weights': 'float',
                               'WeightsV2': 'float',
                               'WeightsV3': 'float',
                               # 'shapetype': 'str',
                               'emoji': 'int'},
            }
        else:
            # Define a polygon feature geometry with one attribute
            schema = {
                'geometry': 'Point',
                'properties': {'Join_Count': 'int',
                               'Weights': 'float'},
            }
        return schema

    @staticmethod
    def _contains_emoji_output(shapes_and_meta_list):
        """Check if emoji type is in output list"""
        contains_emoji_output = False
        for __, output_type in shapes_and_meta_list:
            if output_type == EMOJI:
                contains_emoji_output = True
        return contains_emoji_output

    @classmethod
    def _getcompile_nonitemized_shapes(
            cls, shapes,
            weights: Dict[int, Tuple[float, float]],
    ):
        """Compilation of final records to be

        written to shapefile. Includes normalization
        of values"""
        shapelist = list()
        for alphashape_and_meta in shapes:
            local_value = alphashape_and_meta[1]
            local_value_normalized = cls._get_normalize_value(
                local_value, weights.get(1))
            # each part of the tuple is
            # one column in output shapefile
            shapelist.append(
                (alphashape_and_meta[0],
                 alphashape_and_meta[1],
                 local_value_normalized))
        return shapelist

    @classmethod
    def _write_all(cls, shapes, itemized,
                   contains_emoji_output, epsg_code, output_folder):
        schema = cls._get_shape_schema(itemized)
        # update for emoji only run
        if itemized:
            shapefile_name = "allTagCluster"
            # sort shapelist by firstg column,
            # in descending order
            # we want most important tags places first
            shapes.sort(key=itemgetter(7), reverse=True)
        else:
            shapefile_name = "allLocationCluster"
            # sort ascending, we want smalles clusters places
            # first as small points, overlayed by larger ones
            shapes.sort(key=itemgetter(2))
        with fiona.open(
                output_folder / f'{shapefile_name}.shp', mode='w',
                encoding='UTF-8', driver='ESRI Shapefile',
                schema=schema, crs=from_epsg(epsg_code)) as shapefile:
            cls._attach_emojitable_handler(
                shapefile,
                shapes,
                contains_emoji_output,
                itemized, output_folder)

    @classmethod
    def _attach_emojitable_handler(cls, shapefile,
                                   shapes,
                                   contains_emoji_output,
                                   itemized, output_folder):
        """If Emoji Output present, open csv for writing
        Note: refactor as optional property!
        """
        if contains_emoji_output:
            # If newline is '', no translation takes place on write
            # that means: \n (LF) is written, not CRLF
            with open(output_folder / f'emojiTable.csv',
                      "w", newline='', encoding='utf-8') as emoji_table:
                emoji_table.write("FID,Emoji\n")
                if itemized:
                    cls._write_all_shapes(
                        shapefile, shapes,
                        emoji_table, itemized)

        else:
            cls._write_all_shapes(
                shapefile, shapes, None, itemized)

    @classmethod
    def _write_all_shapes(cls, shapefile, shapes,
                          emoji_table: TextIO,
                          itemized: bool):
        fid = 0
        for shape in shapes:
            if itemized:
                # check if colum 10 is set to 1
                # == emoji record
                is_emoji_record = bool(shape[10] == 1)
                cls._write_itemized_shape(
                    shapefile, shape, is_emoji_record)
                if emoji_table:
                    cls._write_emoji_record(
                        fid,
                        shape,
                        emoji_table,
                        is_emoji_record)
                    fid += 1
            else:
                cls._write_nonitemized_shape(
                    shapefile, shape)

    @staticmethod
    def _write_nonitemized_shape(shapefile, shape):
        shapefile.write({
            'geometry': geometry.mapping(shape[0]),
            'properties': {'Join_Count': shape[1],
                           'Weights': shape[2]},
        })

    @staticmethod
    def _write_itemized_shape(shapefile, shape, is_emoji_record):
        """Append final record to shape"""
        # do not write emoji to shapefile directly
        # bug in Arcgis, needs to be imported
        # using join
        if is_emoji_record:
            imptag = ""
        else:
            imptag = shape[4]
        shapefile.write({
            'geometry': geometry.mapping(shape[0]),
            'properties': {'Join_Count': shape[1],
                           'Views': shape[2],
                           'COUNT_User': shape[3],
                           'ImpTag': imptag,
                           'TagCountG': shape[5],
                           'HImpTag': shape[6],
                           'Weights': shape[7],
                           'WeightsV2': shape[8],
                           'WeightsV3': shape[9],
                           # 'shapetype': alphaShapeAndMeta[9],
                           'emoji': shape[10]},
        })

    @classmethod
    def _getcompile_itemized_shapes(
            cls, shapes, cls_type,
            weights: Dict[int, Tuple[float, float]]):
        """Main wrapper for writing
        all results to output
        """
        # Normalize Weights to 0-1000 Range
        idx = 0
        shapelist = list()
        for alphashape_and_meta in shapes:
            h_imp = cls._get_himp(idx, shapes)
            idx += 1
            item_shape = cls._getcompile_item_shape(
                alphashape_and_meta,
                cls_type, weights, h_imp)
            shapelist.append((item_shape))
        return shapelist

    @staticmethod
    def _get_himp(idx, shapes):
        """check if current cluster is most
        used for item
        Compares item str to previous item,
        if different, then himp=1
        Note: Items are ordered,
        therefore a change means
        new item begins
        """
        if ((idx == 0
             or shapes[idx][4] != shapes[idx-1][4])
                and not shapes[idx][2] == 1):
            # if item is different to
            # previous item and
            # user count is not 1
            h_imp = 1
        else:
            h_imp = 0
        return h_imp

    @staticmethod
    def _write_emoji_record(fid, shape, emoji_table,
                            is_emoji):
        """Write Emoji table separetely to join to shapefile"""
        if is_emoji:
            imp_tag_text = f'{shape[4]}'
        else:
            # also write tags as empty
            # records to table, necessary
            # for accurate join/ fid count
            imp_tag_text = ""
        emoji_table.write(
            f'{fid},{imp_tag_text}\n')

    @classmethod
    def _getcompile_item_shape(
            cls,
            alphashape_and_meta, cls_type,
            weights: Dict[int, Tuple[float, float]],
            h_imp):

        # emoName = unicode_name(alphaShapeAndMeta[4])
        # Calculate Normalized Weights Values based on precalc Step
        value_weight1 = alphashape_and_meta[6]
        weight1_normalized = cls._get_normalize_value(
            value_weight1, weights.get(1))
        value_weight2 = alphashape_and_meta[7]
        weight2_normalized = cls._get_normalize_value(
            value_weight2, weights.get(2))
        value_weight3 = alphashape_and_meta[8]
        weight3_normalized = cls._get_normalize_value(
            value_weight3, weights.get(3))
        # project data
        # geom_proj = transform(project, alphaShapeAndMeta[0])
        # c.write({
        #    'geometry': geometry.mapping(geom_proj),
        if cls_type == EMOJI:
            emoji = 1
            # due to bug in ArcGIS
            # leave blank
            # emoji must be imported separately
            # imp_tag_text = ""
        else:
            emoji = 0
            # imp_tag_text = f'{alphashape_and_meta[4]}'
        item_shape = (
            alphashape_and_meta[0],
            alphashape_and_meta[1],
            alphashape_and_meta[2],
            alphashape_and_meta[3],
            alphashape_and_meta[4],
            alphashape_and_meta[5],
            h_imp,
            weight1_normalized,
            weight2_normalized,
            weight3_normalized,
            emoji
        )
        return item_shape

    @classmethod
    def _get_weights(cls, shapes, columns: List[int]):
        """Normalization of Values (1-1000 Range),
        precalc Step (global weights),
        see
        https://stats.stackexchange.com/questions/70801/how-to-normalize-data-to-0-1-range
        """
        idx = 1
        weights_dict: Dict[int, Tuple[float, float]] = dict()
        for column in columns:
            # int: weights algorithm
            # Tuple: min and max modifiers
            values_min, values_max = cls._get_column_min_max(
                shapes, column)
            weights_mod_a = (1000-1)/(
                values_max-values_min)
            weights_mod_b = 1000 - weights_mod_a * values_max
            weights_dict[idx] = (weights_mod_a, weights_mod_b)
            idx += 1
        cls._write_legend_info(weights_dict)
        return weights_dict

    @staticmethod
    def _write_legend_info(weights_dict):
        """Helper Function for updating ArcGIS legend
        - not implemented
        """
        if not weights_dict:
            return
        return

    @staticmethod
    def _get_column_min_max(
            shapes,
            column: int) -> Tuple[float, float]:
        """Get min and max values of specific column
        for normalization"""
        # get the n'th column out for calculating the max/min
        weights_range = [x[column] for x in shapes]
        weights_min = min(weights_range)
        weights_max = max(weights_range)
        return weights_min, weights_max

    @staticmethod
    def _normalize_value(
            weights,
            local_value):
        """Normalize value based on global weights"""
        mul_mod = weights[0]
        sum_mod = weights[1]
        normalized_value = mul_mod * local_value + sum_mod
        return normalized_value

    @classmethod
    def _get_normalize_value(cls, local_value,
                             weights_tuple: Tuple[float, float]):
        """Wrapper for Normalization: keep 1,
        otherwise, normalize"""
        if local_value == 1:
            value_normalized = local_value
        else:
            value_normalized = cls._normalize_value(
                weights_tuple, local_value)
        return value_normalized

    @staticmethod
    def get_weight(weighting_id: int,
                   post_count: int,
                   unique_user_count: int):
        """Get Weight for input metrics
        user count and post count based on
        three types of weighting formulas (id):
        1: weightsv1 -> Standard weighting formula
           (x**y means x raised to the power y);
           +1 to UserCount: prevent 1-2
           Range from being misaligned
        2: weightsv2 -> less importance
           on User_Count compared
           to photo count [Join_Count];
           +1 to UserCount: prevent 1-2
           Range from being misaligned
        3: weightsv3 -> Ignores User_Count,
           this will emphasize individual
           and very active users
        """
        if weighting_id == 1:
            weighted_value = post_count * (
                sqrt(1/(post_count / (
                    unique_user_count+1))**3))
        elif weighting_id == 2:
            weighted_value = post_count * (
                sqrt(1/(post_count / (
                    unique_user_count+1))**2))
        elif weighting_id == 3:
            weighted_value = sqrt(
                (post_count+(2*sqrt(post_count)))*2)
        return weighted_value
